import json
import socket
from datetime import timedelta

import torch.distributed as dist
import torch
import os

from ..logger import beam_logger as logger
from ..core import Processor
from ..utils import has_kwargs, GPUManager


class BeamFederated(Processor):
    def __init__(self, *args, func=None, rank=0, world_size=1, framework='ddp', distributed_backend='nccl', host=None,
                 port=None, func_args=None, func_kwargs=None, kv_store='tcp', kv_store_path=None,
                 kv_store_timeout=300, kv_store_port=None, devices=None, **kwargs):

        super().__init__(*args, training_framework=framework, mp_port=port, mp_ip=host,
                         kv_store_path=kv_store_path, kv_store_timeout=kv_store_timeout, kv_store_port=kv_store_port,
                         distributed_backend=distributed_backend, kv_store=kv_store, **kwargs)

        self.rank = rank
        self.world_size = world_size

        self.func = func
        self.func_args = func_args if func_args is not None else []
        self.func_kwargs = func_kwargs if func_kwargs is not None else {}

        self.host = self.get_hparam('mp_ip') or 'localhost'

        self.port = str(self.get_hparam('mp_port'))
        self.backend = self.get_hparam('distributed_backend')
        self.framework = self.get_hparam('training_framework')

        if devices is None:
            devices = list(range(torch.cuda.device_count()))
        else:
            if not isinstance(devices, list):
                devices = [int(devices)]
            else:
                devices = [int(d) for d in devices]
        self.devices = devices

        self.kv_store = self.get_hparam('kv_store')
        self.kv_store_path = self.get_hparam('kv_store_path')
        self.kv_store_timeout = self.get_hparam('kv_store_timeout')
        self.kv_store_port = self.get_hparam('kv_store_port')

        logger.info(f"Rank {self.rank}/{self.world_size} connects to the KV store: {self.kv_store}")
        if self.kv_store == 'tcp':
            self.store = dist.TCPStore(host_name=self.host, port=int(self.kv_store_port), world_size=self.world_size,
                                       is_master=(self.rank == 0), timeout=timedelta(seconds=self.kv_store_timeout))
        elif self.kv_store == 'hash':
            self.store = dist.HashStore()
        elif self.kv_store == 'file':
            self.store = dist.FileStore(self.kv_store_path, self.world_size)
        else:
            raise ValueError(f"Unknown kv_store: {self.kv_store}")

        if self.world_size > 1:
            self._init_distributed()

    @property
    def physical_devices(self):
        return GPUManager.physical_devices(self.devices)

    @property
    def hostname(self):
        return socket.gethostname()

    def _init_distributed(self):
        os.environ['MASTER_ADDR'] = self.host
        os.environ['MASTER_PORT'] = self.port

        logger.info(f"Rank {self.rank}/{self.world_size} is initializing distributed training (host={self.host}, port={self.port})")
        logger.info(f"Initializing distributed training with backend={self.backend} and framework={self.framework}")
        if self.framework == 'ddp':
            # initialize the process group
            dist.init_process_group(self.backend, rank=self.rank, world_size=self.world_size, store=self.store)
        elif self.framework == 'deepspeed':

            # make sure that mpi path is in the path variable

            os.environ['LOCAL_RANK'] = str(self.rank)
            os.environ['RANK'] = str(self.rank)
            os.environ['WORLD_SIZE'] = str(self.world_size)
            os.environ['OMPI_COMM_WORLD_SIZE'] = str(self.world_size)
            os.environ['OMPI_COMM_WORLD_RANK'] = str(self.rank)

            import deepspeed
            # deepspeed.init_distributed(dist_backend=backend, auto_mpi_discovery=backend == 'mpi',
            #                            rank=rank, world_size=world_size, distributed_port=port)
            deepspeed.init_distributed(dist_backend=self.backend, auto_mpi_discovery=False,
                                       rank=self.rank, world_size=self.world_size, distributed_port=self.port)

        else:
            raise ValueError(f"Unknown distributed framework: {self.framework}")

    def barrier(self, group=None, async_op=None, device_ids=None):
        kwargs = {}
        if group is not None:
            kwargs['group'] = group
        if async_op is not None:
            kwargs['async_op'] = async_op
        if device_ids is not None:
            kwargs['device_ids'] = device_ids
        if self.framework == 'ddp':
            import torch.distributed as dist
            dist.barrier(**kwargs)
        elif self.framework == 'deepspeed':
            import deepspeed
            deepspeed.comm.barrier(**kwargs)
        else:
            raise ValueError(f"Unknown distributed framework: {self.framework}")

    def __getitem__(self, item):
        value = self.store.get(item)
        return json.loads(value)

    def __setitem__(self, key, value):
        value = json.dumps(value)
        return self.store.set(key, value)

    def __delitem__(self, key):
        return self.store.delete_key(key)

    def add_to_key(self, key, value):
        return self.store.add(key, value)

    def num_keys(self):
        return self.store.num_keys()

    def wait(self, keys):
        if isinstance(keys, str):
            keys = [keys]
        self.store.wait(keys)

    def __call__(self, *args, func=None, **kwargs):
        if func is None:
            func = self.func

        args = list(args) + self.func_args
        kwargs = {**kwargs, **self.func_kwargs}
        kwargs['manager'] = self

        return func(*args, **kwargs)

    def cleanup(self):
        if self.world_size > 1:
            if self.framework == 'ddp':
                dist.destroy_process_group()
            elif self.framework == 'deepspeed':
                pass
            else:
                raise ValueError(f"Unknown distributed framework: {self.framework}")


# For MPI use the spinningup resource:
# https://github.com/openai/spinningup/blob/038665d62d569055401d91856abb287263096178/spinup/utils/mpi_tools.py#L4


import copy
from queue import Queue, Empty
from threading import Thread
from flask import Flask, request, jsonify, send_file

from ..logger import beam_logger as logger
from .server import BeamServer

try:
    import torch
    has_torch = True
except ImportError:
    has_torch = False


class HTTPServer(BeamServer):
    """
    Class representing a server for executing inference on an object or function.

    Args:
        obj: The object or function to perform inference on.
        use_torch (bool): Whether to use torch for serialization and deserialization. Default is True.
        batch: The methods to batch execute. Can be a boolean, string or list.
               If True, it will batch execute all the methods with '__call__' in the name.
               If a string, it will batch execute the method with the specified name.
               If a list, it will batch execute the methods with the names in the list.
        max_wait_time (float): The maximum time to wait for new inference tasks in a batch. Default is 1.0 second.
        max_batch_size (int): The maximum size of a batch for batch inference. Default is 10.
        tls (bool): Whether to use a secure TLS connection. Default is False.
        n_threads (int): The number of threads to use for the server. Default is 4.
        **kwargs: Additional keyword arguments.

    Attributes:
        app (Flask): The Flask application object.
        obj: The object or function being served.
        load_function: The function used for deserializing data.
        dump_function: The function used for serializing data.
        serialization_method (str): The serialization method used ('torch' or 'pickle').
        max_wait_time (float): The maximum time to wait for new inference tasks in a batch.
        max_batch_size (int): The maximum size of a batch for batch inference.
        tls (bool): Whether to use a secure TLS connection.
        n_threads (int): The number of threads used for the server.
        _request_queue (Queue): The queue for incoming inference tasks.
        _response_queue (defaultdict(Queue)): The queue for outgoing inference results.
        centralized_thread (Thread): The thread used for centralized batch inference.

    Methods:
        __init__: Initializes the BeamServer object.
        set_variable: Sets the value of a variable in the object being served.
        get_variable: Gets the value of a variable from the object being served.
        _cleanup: Cleans up resources used by the server.
        request_queue: Returns the request queue, creating it if necessary.
        response_queue: Return the response queue, creating it if necessary.
        build_algorithm_from_path: Creates a BeamServer object from an experiment file.
        run_non_blocking: Starts the server in a separate thread.
        run: Starts the server on the specified host and port.
        _centralized_batch_executor: Executes batch inference in a centralized manner.
        get_info: Gets information about the server and the object being served.
        batched_query_algorithm: Executes a method on the object being served in batch mode.
        call_function: Executes the '__call__' method on the function being served.
        query_algorithm: Executes a method on the object being served.
        run_uwsgi: Starts the server using uWSGI.
        run_waitress: Starts the server using Waitress.
        run_cherrypy: Starts the server using CherryPy.
        run_gunicorn: Starts the server using Gunicorn.
        run_wsgi: Starts the server using WSGI.
    """
    def __init__(self, *args, application=None, **kwargs):
        application = 'flask' or application
        super().__init__(*args, application=application, **kwargs)
        self.app = Flask(__name__)
        self.app.add_url_rule('/', view_func=self.get_info)

        if self.type == 'function':
            self.app.add_url_rule('/call/<client>', view_func=self.call, methods=['POST'])
        elif self.type == 'instance':
            if hasattr(self.obj, '__call__'):
                self.app.add_url_rule('/call/<client>', view_func=self.call, methods=['POST'])
            self.app.add_url_rule('/alg/<client>/<method>', view_func=self.query_algorithm, methods=['POST'])
            self.app.add_url_rule('/setvar/<client>/<name>', view_func=self.set_variable, methods=['POST'])
            self.app.add_url_rule('/getvar/<client>/<name>', view_func=self.get_variable)
        else:
            raise ValueError(f"Unknown type: {self.type}")

    def set_variable(self, client, name, *args, **kwargs):
        value = request.files['value']
        res = super().set_variable(client, name, value)
        return jsonify(res)

    def get_variable(self, client, name):
        io_results = super().get_variable(client, name)
        if client == 'beam':
            return send_file(io_results, mimetype="text/plain")
        else:
            return jsonify(io_results)

    def _run(self, host="0.0.0.0", port=None, server='waitress', use_reloader=True):

        # when debugging with pycharm set debug=False
        # if needed set use_reloader=False
        # see https://stackoverflow.com/questions/25504149/why-does-running-the-flask-dev-server-run-itself-twice

        if port is not None:
            port = int(port)

        if server == 'debug':
            self.app.run(host=host, port=port, debug=True, use_reloader=use_reloader, threaded=True)
        elif server == 'wsgi':
            self.run_wsgi(host, port)
        elif server == 'uwsgi':
            self.run_uwsgi(host, port)
        elif server == 'waitress':
            self.run_waitress(host, port)
        elif server == 'cherrypy':
            self.run_cherrypy(host, port)
        elif server == 'gunicorn':
            self.run_gunicorn(host, port)
        else:
            raise ValueError(f"Unknown serve type: {server}")

    def get_info(self):
        d = super().get_info()
        return jsonify(d)

    @staticmethod
    def request_metadata(client='beam', method=None):

        metadata = {
            'method': request.method,
            'url': request.url,
            'base_url': request.base_url,
            'args': copy.deepcopy(request.args),
            'form': copy.deepcopy(request.form),
            'headers': copy.deepcopy(dict(request.headers)),
            'remote_addr': request.remote_addr,
            'user_agent': str(request.user_agent),
            'client': client,
            'alg_method': method,

        }

        return metadata

    def query_algorithm(self, client, method, *args, **kwargs):

        if client == 'beam':
            args = request.files['args']
            kwargs = request.files['kwargs']
        else:
            data = request.get_json()
            args = data.pop('args', [])
            kwargs = data.pop('kwargs', {})

        io_results = super().query_algorithm(client, method, args, kwargs)

        if client == 'beam':
            return send_file(io_results, mimetype="text/plain")
        else:
            return jsonify(io_results)

    def run_uwsgi(self, host, port):

        from uwsgi import run

        uwsgi_opts = {
            'http': f'{host}:{port}',
            'wsgi-file': 'your_wsgi_file.py',  # Replace with your WSGI file
            'callable': 'app',  # Replace with your WSGI application callable
            'processes': self.n_threads,
        }

        if self.tls:
            uwsgi_opts['https-socket'] = f'{host}:{port}'
            uwsgi_opts['https-keyfile'] = 'path/to/keyfile.pem'
            uwsgi_opts['https-certfile'] = 'path/to/certfile.pem'

        run([], **uwsgi_opts)

    def run_waitress(self, host, port):

        from waitress import serve
        serve(self.app, host=host, port=port, threads=self.n_threads)

    def run_cherrypy(self, host, port):

        import cherrypy

        cherrypy.tree.graft(self.app, '/')
        config = {
            'serve.socket_host': host,
            'serve.socket_port': port,
            'engine.autoreload.on': False,
            'serve.thread_pool': self.n_threads
        }
        if self.tls:
            config.update({
                'serve.ssl_module': 'builtin',
                'serve.ssl_certificate': 'path/to/certfile.pem',
                'serve.ssl_private_key': 'path/to/keyfile.pem'
            })
        cherrypy.config.update(config)

        cherrypy.engine.start()
        cherrypy.engine.block()

    def run_gunicorn(self, host, port):

        from gunicorn.app.wsgiapp import WSGIApplication
        options = {
            'bind': f'{host}:{port}',
            'workers': 1,  # Gunicorn forks multiple processes and is generally not thread-safe
            'threads': self.n_threads,
            'accesslog': '-',
        }

        if self.tls:
            options['keyfile'] = 'path/to/keyfile.pem'
            options['certfile'] = 'path/to/certfile.pem'

        app = WSGIApplication()
        app.load_wsgiapp = lambda: self.app
        app.cfg.set(options)
        app.run()

    def run_wsgi(self, host, port):

        from gevent.pywsgi import WSGIServer

        if self.n_threads > 1:
            logger.warning("WSGI serve does not support multithreading, setting n_threads to 1")

        if self.tls:
            from gevent.pywsgi import WSGIServer
            from gevent.ssl import SSLContext
            from os.path import join, dirname, realpath

            cert = join(dirname(realpath(__file__)), 'cert.pem')
            key = join(dirname(realpath(__file__)), 'key.pem')
            context = SSLContext()
            context.load_cert_chain(cert, key)
        else:
            context = None

        http_server = WSGIServer((host, port), self.app, ssl_context=context)
        http_server.serve_forever()



# setup.py
import importlib


class BeamImporter:
    def __init__(self):
        self.modules = {}
        self.callables = {}
        self.aliases = {
            'pd': 'pandas',
            'np': 'numpy',
            'torch': 'torch',
            'F': 'torch.nn.functional',
            'Path': 'pathlib.Path',
            'plt': 'matplotlib.pyplot',
            # 'sns': 'seaborn',
            # 'tg': 'torch_geometric',
            'nn': 'torch.nn',
            'optim': 'torch.optim',
            'distributions': 'torch.distributions',
            'os': 'os',
            'sys': 'sys',
            'inspect': 'inspect',
            'time': 'time',
            'timedelta': 'datetime.timedelta',
            'random': 'random',
            'nx': 'networkx',
            're': 're',
            'glob': 'glob',
            'pickle': 'pickle',
            'json': 'json',
            'datetime': 'datetime.datetime',
            'date': 'datetime.date',
            'tqdm': 'tqdm.notebook.tqdm',
            'beam': 'beam',
            'defaultdict': 'collections.defaultdict',
            'Counter': 'collections.Counter',
            'OrderedDict': 'collections.OrderedDict',
            'partial': 'functools.partial',
            'namedtuple': 'collections.namedtuple',
            'reduce': 'functools.reduce',
            'itertools': 'itertools',
            'copy': 'copy',
            'warnings': 'warnings',
            'deepcopy': 'copy.deepcopy',


        }
        self.__all__ = list(self.aliases.keys())  # Iterable of attribute names
        self.__file__ = None
        self._initialize_aliases()

    def _initialize_aliases(self):
        pass

    def __getattr__(self, name):
        if name in self.aliases:
            actual_name = self.aliases[name]
        else:
            actual_name = name

        try:
            imported_object = importlib.import_module(actual_name)
            # print(f"Info: actual_name: {actual_name}, imported_object: {imported_object}")
        except Exception as e:
            try:
                module_name, object_name = actual_name.rsplit('.', 1)
                module = importlib.import_module(module_name)
                imported_object = getattr(module, object_name)
                # print(f"Info: module_name: {module_name}, object_name: {object_name}, imported_object: {imported_object}")
            except:
                print(f"Warning: Could not import {actual_name}, {e}, Skipping...")
                imported_object = None

        return imported_object


def load_ipython_extension(ipython):
    import sys
    import os
    import time

    t0 = time.time()
    os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'

    print("Setting up the Beam environment for interactive use")
    print("Standard modules will be automatically imported so you can use them without explicit import")

    if ipython is not None:
        ipython.run_line_magic('load_ext', 'autoreload')
        ipython.run_line_magic('autoreload', '2')

    beam_path = os.getenv('BEAM_PATH', None)
    if beam_path is not None:
        sys.path.insert(0, beam_path)
        sys.path.insert(0, os.path.join(beam_path, 'src'))
    else:
        sys.path.insert(0, '..')
        sys.path.insert(0, '../src')

    beam_importer = BeamImporter()

    # Add the modules to the global namespace
    for alias in beam_importer.aliases:
        module = getattr(beam_importer, alias)
        if ipython is not None:
            ipython.push({alias: module})
        if alias == 'beam':
            for k in module.__dict__.keys():
                if not k.startswith('_') and k in module.__all__:
                    ipython.push({k: module.__dict__[k]})

    print(f"Done importing packages. It took: {time.time() - t0: .2} seconds")
    print(f"Beam library is loaded from path: {os.path.abspath(os.path.dirname(beam_importer.beam.__file__))}")
    print(f"The Beam version is: {beam_importer.beam.__version__}")


if __name__ == '__main__':
    load_ipython_extension(None)


from ..utils import (divide_chunks, Timer, collate_chunks, retrieve_name,
                     jupyter_like_traceback, lazy_property, dict_to_list)
from ..logger import beam_logger as logger


class TaskResult:
    def __init__(self, async_result):

        from celery.result import AsyncResult as CeleryAsyncResult
        from multiprocessing.pool import AsyncResult as MultiprocessingAsyncResult

        self.async_result = async_result
        if isinstance(async_result, CeleryAsyncResult):
            self.method = 'celery'
        elif isinstance(async_result, MultiprocessingAsyncResult):
            self.method = 'apply_async'
        else:
            raise ValueError(
                "Invalid async_result type. It must be either CeleryAsyncResult or MultiprocessingAsyncResult.")

    @property
    def done(self):
        if self.method == 'celery':
            return self.async_result.ready()
        else:  # method == 'apply_async'
            return self.async_result.ready()

    @property
    def result(self):

        if self.method == 'celery':
            return self.async_result.result if self.done else None
        else:  # method == 'apply_async'
            return self.async_result.get() if self.done else None


class SyncedResults:

    def __init__(self, results):
        # results is a list of dicts with keys: name, result, exception
        self.results = results

    @lazy_property
    def results_map(self):
        return {r['name']: r for r in self.results}

    @lazy_property
    def failed(self):
        failed = {r['name']: r['exception'] for r in self.results if r['exception'] is not None}
        return dict_to_list(failed, convert_str=False)

    @lazy_property
    def succeeded(self):
        succeeded = {r['name']: r['result'] for r in self.results if r['exception'] is None}
        return dict_to_list(succeeded, convert_str=False)

    @lazy_property
    def values(self):
        vals = {r['name']: r['result'] if r['exception'] is None else r['exception'] for r in self.results}
        return dict_to_list(vals, convert_str=False)

    @lazy_property
    def exceptions(self):
        vals = {r['name']: r['exception'] for r in self.results if r['exception'] is not None}
        return dict_to_list(vals, convert_str=False)


class BeamTask(object):

    def __init__(self, func, *args, name=None, silence=False, metadata=None, **kwargs):

        self.func = func
        self.args = args
        self.kwargs = kwargs
        self._name = name
        self.pid = None
        self.is_pending = True
        self.result = None
        self.exception = None
        self.metadata = metadata
        self.queue_id = -1
        self.silence = silence

    @property
    def name(self):
        if self._name is None:
            self._name = retrieve_name(self)
        return self._name

    def set_silent(self, silence):
        self.silence = silence

    def set_name(self, name):
        self._name = name

    def __call__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

    def run(self):

        metadata = f"({self.metadata})" if self.metadata is not None else ""

        if not self.silence:
            logger.info(f"Starting task: {self.name} {metadata}")
        try:
            with Timer(logger, silence=True) as t:
                res = self.func(*self.args, **self.kwargs)
                self.result = res
                if not self.silence:
                    logger.info(f"Finished task: {self.name} {metadata}. Elapsed time: {t.elapsed}")
        except Exception as e:
            self.exception = e
            logger.error(f"Task {self.name}{metadata} failed with exception: {e}")
            res = jupyter_like_traceback()
        finally:
            self.is_pending = False

        return {'name': self.name, 'result': res, 'exception': self.exception}



import fnmatch
import json
import os
from collections import namedtuple
from datetime import datetime
from io import BytesIO, StringIO
from pathlib import PurePosixPath, PureWindowsPath, Path
from typing import Union
from urllib.parse import urlparse, urlunparse, parse_qsl, ParseResult
import pandas as pd
import numpy as np
import re

BeamFile = namedtuple('BeamFile', ['data', 'timestamp'])


def strip_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


class BeamURL:

    def __init__(self, url=None, scheme=None, hostname=None, port=None, username=None, password=None, path=None,
                 fragment=None, params=None, **query):

        self._url = url
        self._parsed_url = None
        if url is None:
            netloc = BeamURL.to_netloc(hostname=hostname, port=port, username=username, password=password)
            query = BeamURL.dict_to_query(**query)
            if scheme is None:
                scheme = 'file'
            if path is None:
                path = ''
            if netloc is None:
                netloc = ''
            if query is None:
                query = ''
            if fragment is None:
                fragment = ''
            if params is None:
                params = ''
            self._parsed_url = ParseResult(scheme=scheme, netloc=netloc, path=path, params=params, query=query,
                                           fragment=fragment)

        assert self._url is not None or self._parsed_url is not None, 'Either url or parsed_url must be provided'

    @property
    def parsed_url(self):
        if self._parsed_url is not None:
            return self._parsed_url
        self._parsed_url = urlparse(self._url)
        return self._parsed_url

    @property
    def url(self):
        if self._url is not None:
            return self._url
        self._url = urlunparse(self._parsed_url)
        return self._url

    def __repr__(self):
        return self.url

    def __str__(self):

        netloc = BeamURL.to_netloc(hostname=self.hostname, port=self.port, username=self.username)
        parsed_url = ParseResult(scheme=self.scheme, netloc=netloc, path=self.path, params=None, query=None,
                                 fragment=None)
        return urlunparse(parsed_url)

    @property
    def scheme(self):
        return self.parsed_url.scheme

    @property
    def protocol(self):
        return self.scheme

    @property
    def username(self):
        return self.parsed_url.username

    @property
    def hostname(self):
        return self.parsed_url.hostname

    @property
    def password(self):
        return self.parsed_url.password

    @property
    def port(self):
        return self.parsed_url.port

    @property
    def path(self):
        return self.parsed_url.path

    @property
    def query_string(self):
        return self.parsed_url.query

    @property
    def query(self):
        return dict(parse_qsl(self.parsed_url.query))

    @property
    def fragment(self):
        return self.parsed_url.fragment

    @property
    def params(self):
        return self.parsed_url.params

    @staticmethod
    def to_netloc(hostname=None, port=None, username=None, password=None):

        if not hostname:
            return None

        netloc = hostname
        if username:
            if password:
                username = f"{username}:{password}"
            netloc = f"{username}@{netloc}"
        if port:
            netloc = f"{netloc}:{port}"
        return netloc

    @staticmethod
    def to_path(path):
        return PurePosixPath(path).as_posix()

    @staticmethod
    def query_to_dict(query):
        return dict(parse_qsl(query))

    @staticmethod
    def dict_to_query(**query):
        return '&'.join([f'{k}={v}' for k, v in query.items() if v is not None])

    @classmethod
    def from_string(cls, url):
        parsed_url = urlparse(url)
        return cls(url, parsed_url)


def normalize_host(hostname, port=None, default='localhost'):
    if hostname is None:
        hostname = default
    if port is None:
        host = f"{hostname}"
    else:
        host = f"{hostname}:{port}"

    return host


class BeamResource:
    """
    Base class for all resources. Gets as an input a URI and the resource type and returns the resource.
    """

    def __init__(self, resource_type: str, url: Union[BeamURL, str] = None, scheme: str = None,
                 hostname: str = None, port: int = None, username: str = None, password: str = None,
                 fragment: str = None, params: str = None, path: str = None, **kwargs):

        if isinstance(url, str):
            url = BeamURL(url)

        if url is not None:
            scheme = scheme or url.scheme
            hostname = hostname or url.hostname
            port = port or url.port
            username = username or url.username
            password = password or url.password
            fragment = fragment or url.fragment
            params = params or url.params
            kwargs = kwargs or url.query
            path = path or url.path

        self.url = BeamURL(scheme=scheme, hostname=hostname, port=port, username=username, fragment=fragment,
                           params=params, password=password, path=path, **kwargs)

        self.resource_type = resource_type
        self.scheme = self.url.scheme

    def as_uri(self):
        return self.url.url

    @property
    def hostname(self):
        return self.url.hostname

    @property
    def port(self):
        return self.url.port

    @property
    def username(self):
        return self.url.username

    @property
    def password(self):
        return self.url.password

    @property
    def fragment(self):
        return self.url.fragment

    @property
    def params(self):
        return self.url.params

    @property
    def query(self):
        return self.url.query


class PureBeamPath(BeamResource):
    feather_index_mark = "feather_index:"

    # all the extensions that are considered textual (should be read as .txt)
    textual_extensions = ['.txt', '.text', '.py', '.sh', '.c', '.cpp', '.h', '.hpp', '.java', '.js', '.css',
                          '.html', '.md']
    text_based_extensions = textual_extensions + ['.json', '.orc', '.yaml', '.yml', '.ndjson', '.csv', '.ini']

    def __init__(self, *pathsegments, scheme=None, client=None, **kwargs):

        if len(pathsegments) == 1 and isinstance(pathsegments[0], PureBeamPath):
            pathsegments = pathsegments[0].parts

        if scheme == 'nt':
            self.path = PureWindowsPath(*pathsegments)
        else:
            self.path = PurePosixPath(*pathsegments)

        super().__init__(resource_type='storage', scheme=scheme, path=str(self.path), **kwargs)

        self.mode = "rb"
        self.file_object = None
        self.close_fo_after_read = None
        self.client = client
        self.open_kwargs = dict(mode="rb", buffering=- 1, encoding=None, errors=None,
                                newline=None, closefd=True, opener=None)

    @property
    def str(self):
        return str(self.path)

    @property
    def list(self):
        return list(self)

    def __getstate__(self):
        return self.as_uri()

    def __setstate__(self, state):

        url = BeamURL.from_string(state)

        self.__init__(url.path, hostname=url.hostname, port=url.port, username=url.username,
                      password=url.password, fragment=url.fragment, params=url.params, client=None, **url.query)

    def __iter__(self):
        for p in self.iterdir():
            yield p

    def __getitem__(self, name):
        return self.joinpath(name)

    def __setitem__(self, key, value):
        p = self.joinpath(key)
        p.write(value)

    def touch(self, mode=0o666, exist_ok=True):
        self.write(b'', ext='.bin')

    def not_empty(self, filter_pattern=None):

        if self.is_dir():
            for p in self.iterdir():
                if p.not_empty():
                    return True
                if p.is_file():
                    if filter_pattern is not None:
                        if not re.match(filter_pattern, p.name):
                            return True
                    else:
                        return True
        return False

    def copy(self, dst, ignore=None, include=None):

        if type(dst) is str:
            dst = self.gen(dst)

        if self.is_dir():
            dst.mkdir(parents=True, exist_ok=True)
            for p in self.iterdir():
                p.copy(dst.joinpath(p.name), ignore=ignore, include=include)
        else:
            dst.parent.mkdir()
            ext = self.suffix
            if ignore is not None:
                if type(ignore) is str:
                    ignore = [ignore]
                if ext in ignore:
                    return
            if include is not None:
                if type(include) is str:
                    include = [include]
                if ext not in include:
                    return
            with self.open("rb") as f:
                with dst.open("wb") as g:
                    g.write(f.read())

    def rmtree(self, ignore=None, include=None):
        if self.is_file():
            self.unlink()
        elif self.is_dir():
            delete_dir = True
            for item in self.iterdir():
                if item.is_dir():
                    item.rmtree(ignore=ignore, include=include)
                else:
                    ext = item.suffix
                    if ignore is not None:
                        if type(ignore) is str:
                            ignore = [ignore]
                        if ext in ignore:
                            delete_dir = False
                            continue
                    if include is not None:
                        if type(include) is str:
                            include = [include]
                        if ext not in include:
                            delete_dir = False
                            continue
                    item.unlink()
            if delete_dir:
                self.rmdir()

    def walk(self):
        dirs = []
        files = []
        for p in self.iterdir():
            if p.is_dir():
                dirs.append(p.name)
            else:
                files.append(p.name)

        yield self, dirs, files

        for dir in dirs:
            yield from self.joinpath(dir).walk()

    def clean(self, ignore=None, include=None):

        if self.exists():
            self.rmtree(ignore=ignore, include=include)
        else:
            if self.parent.exists():
                for p in self.parent.iterdir():
                    if p.stem == self.name:
                        p.rmtree(ignore=ignore, include=include)

        self.mkdir(parents=True)
        self.rmdir()

    @property
    def is_local(self):
        return self.url.scheme == 'file'

    def getmtime(self):
        return None

    def getctime(self):
        return None

    def getatime(self):
        return None

    def stat(self):
        raise NotImplementedError

    def rmdir(self):
        raise NotImplementedError

    def unlink(self, **kwargs):
        raise NotImplementedError

    def __truediv__(self, other):
        return self.joinpath(str(other))

    def __fspath__(self, mode="rb"):
        return str(self)
        # raise TypeError("For BeamPath (named bp), use bp.open(mode) instead of open(bp, mode)")

    def __call__(self, *args, **kwargs):
        return self.open(*args, **kwargs)

    def open(self, mode="rb", buffering=- 1, encoding=None, errors=None, newline=None, closefd=True, opener=None):
        self.mode = mode
        self.open_kwargs = dict(mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline,
                                closefd=closefd, opener=opener)
        return self

    def close_at_exit(self):
        if self.close_fo_after_read or self.mode in ['wb', 'w']:
            self.close()

    def close(self):
        if self.file_object is not None:
            self.file_object.close()
            self.file_object = None

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        if self.is_absolute():
            return str(self.url)
        return str(self.path)

    def __enter__(self):
        raise NotImplementedError

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError

    def __eq__(self, other):

        if type(self) != type(other):
            return False
        p = self.resolve()
        o = other.resolve()

        return p.as_uri() == o.as_uri()

    def gen(self, path):

        PathType = type(self)

        return PathType(path, client=self.client, hostname=self.hostname, port=self.port, username=self.username,
                        password=self.password, fragment=self.fragment, params=self.params, **self.query)

    @property
    def parts(self):
        return self.path.parts

    @property
    def drive(self):
        return self.path.drive

    @property
    def root(self):
        return self.path.root

    def is_root(self):
        return str(self.path) == '/'

    @property
    def anchor(self):
        return self.gen(self.path.anchor)

    @property
    def parents(self):
        return tuple([self.gen(p) for p in self.path.parents])

    @property
    def parent(self):
        return self.gen(self.path.parent)

    @property
    def name(self):
        return self.path.name

    @property
    def suffix(self):
        return self.path.suffix

    @property
    def suffixes(self):
        return self.path.suffixes

    @property
    def stem(self):
        return self.path.stem

    def as_posix(self):
        return self.path.as_posix()

    def is_absolute(self):
        return self.path.is_absolute()

    def is_relative_to(self, *other):
        if len(other) == 1 and isinstance(other[0], PureBeamPath):
            other = str(other[0])
        else:
            other = str(PureBeamPath(*other))
        return self.path.is_relative_to(other)

    def is_reserved(self):
        return self.path.is_reserved()

    def joinpath(self, *other):
        return self.gen(self.path.joinpath(*[str(o) for o in other]))

    def match(self, pattern):
        return self.path.match(pattern)

    def relative_to(self, *other):
        if len(other) == 1 and isinstance(other[0], PureBeamPath):
            other = str(other[0])
        else:
            other = str(PureBeamPath(*other))
        return PureBeamPath(self.path.relative_to(other))

    def with_name(self, name):
        return self.gen(self.path.with_name(name))

    def with_stem(self, stem):
        return self.gen(self.path.with_stem(stem))

    def with_suffix(self, suffix):
        return self.gen(self.path.with_suffix(suffix))

    def glob(self, pattern, case_sensitive=None):
        case_sensitive = case_sensitive or True
        return self._glob_recursive(self, pattern, case_sensitive=case_sensitive)

    def _glob_recursive(self, current_path, pattern, case_sensitive=True):

        if not case_sensitive:
            pattern = pattern.lower()

        for p in current_path.iterdir():

            # If item is a directory and pattern requires recursion
            if p.is_dir() and '**' in pattern:
                # Correctly pass `item` as `current_path` for recursive exploration
                yield from self._glob_recursive(p, pattern, case_sensitive=case_sensitive)
            # Match the item's name against the pattern

            name = str(p.relative_to(self))
            if not case_sensitive:
                name = name.lower()

            if fnmatch.fnmatch(name, pattern):
                yield p

    def rglob(self, *args, **kwargs):
        for path in self.path.rglob(*args, **kwargs):
            yield self.gen(path)

    def absolute(self):
        path = self.path.absolute()
        return self.gen(path)

    def samefile(self, other):
        raise NotImplementedError

    def iterdir(self):
        raise NotImplementedError

    def is_file(self):
        raise NotImplementedError

    def is_dir(self):
        raise NotImplementedError

    def mkdir(self, *args, **kwargs):
        raise NotImplementedError

    def exists(self):
        raise NotImplementedError

    def rename(self, target):
        return NotImplementedError

    def replace(self, target):
        return NotImplementedError

    def read(self, ext=None, **kwargs):

        """

        @param ext:
        @param kwargs:
        @return:

        Supports the following formats:
        - .fea: Feather
        - .csv: CSV
        - .pkl, .pickle: Pickle
        - .npy: Numpy
        - .json: JSON
        - .ndjson: Newline delimited JSON
        - .orc: ORC
        - .txt, '.text', '.py', '.sh', '.c', '.cpp', '.h', '.hpp', '.java', '.js', '.css',
                          '.html', '.md': Text
        - .npz: Numpy zip
        - .scipy_npz: Scipy sparse matrix
        - .flac: Soundfile
        - .parquet: Parquet
        - .pt: PyTorch
        - .xls, .xlsx, .xlsm, .xlsb, .odf, .ods, .odt: Excel
        - .avro: Avro
        - .adjlist, .gexf, .gml, .pajek, .graphml: NetworkX
        - .ini: ConfigParser
        - .h5, .hdf5: HDF5
        - .yaml, .yml: YAML
        - .xml: XML
        - .mat: MAT
        - .zip: ZIP
        - .msgpack: MessagePack
        - .geojson: GeoJSON
        - .wav: Soundfile
        - .joblib, .z, .gz, .bz2, .xz, .lzma: Joblib
        - .safetensors: SafeTensors
        - .bin: Bytes

        """

        self.close_fo_after_read = True
        if ext is None:
            ext = self.suffix

        with self(mode=PureBeamPath.mode('read', ext)) as fo:

            if ext == '.fea':

                import pyarrow as pa
                # x = feather.read_feather(pa.BufferReader(fo.read()), **kwargs)
                x = pd.read_feather(fo, **kwargs)

                c = x.columns
                for ci in c:
                    if PureBeamPath.feather_index_mark in ci:
                        index_name = strip_prefix(ci, PureBeamPath.feather_index_mark)
                        x = x.rename(columns={ci: index_name})
                        x = x.set_index(index_name)
                        break

            elif ext == '.csv':
                x = pd.read_csv(fo, **kwargs)
            elif ext in ['.pkl', '.pickle']:
                x = pd.read_pickle(fo, **kwargs)
            elif ext in ['.npy', '.npz', '.npzc']:
                if ext in ['.npz', '.npzc']:
                    self.close_fo_after_read = False

                x = np.load(fo, allow_pickle=True, **kwargs)
            elif ext in PureBeamPath.textual_extensions:
                if 'readlines' in kwargs and kwargs['readlines']:
                    x = fo.readlines()
                else:
                    x = fo.read()
            elif ext == '.scipy_npz':
                import scipy
                x = scipy.sparse.load_npz(fo, **kwargs)
            elif ext == '.flac':
                import soundfile
                x = soundfile.read(fo, **kwargs)
            elif ext == '.parquet':
                x = pd.read_parquet(fo, **kwargs)
            elif ext == '.pt':
                import torch
                x = torch.load(fo, **kwargs)
            elif ext in ['.xls', '.xlsx', '.xlsm', '.xlsb', '.odf', '.ods', '.odt']:
                x = pd.read_excel(fo, **kwargs)
            elif ext == '.avro':
                x = []
                import fastavro
                for record in fastavro.reader(fo):
                    x.append(record)
            elif ext in ['.adjlist', '.gexf', '.gml', '.pajek', '.graphml']:
                import networkx as nx
                read = getattr(nx, f'read_{ext[1:]}')
                x = read(fo, **kwargs)

            elif ext == '.ini':
                import configparser
                x = configparser.ConfigParser()
                x.read_file(fo)
                x = {section: dict(x.items(section)) for section in x.sections()}

            elif ext in ['.json', '.ndjson']:

                # TODO: add json read with fastavro and shcema
                # x = []
                # with open(path, 'r') as fo:
                #     for record in fastavro.json_reader(fo):
                #         x.append(record)

                nd = ext == '.ndjson'
                try:
                    typ = kwargs.pop('typ', 'native')
                    if (typ == 'native') or nd:
                        x = json.load(fo, **kwargs)
                    else:
                        x = pd.read_json(fo, lines=nd, **kwargs)
                except:
                    fo.seek(0)
                    if nd:
                        x = []
                        for line in fo:
                            x.append(json.loads(line))
                    else:
                        x = json.load(fo)

            elif ext == '.orc':
                import pyarrow as pa
                x = pa.orc.read(fo, **kwargs)

            # HDF5 (.h5, .hdf5)
            elif ext in ['.h5', '.hdf5']:
                import h5py
                with h5py.File(fo, 'r') as f:
                    x = {key: f[key][...] for key in f.keys()}

            # YAML (.yaml, .yml)
            elif ext in ['.yaml', '.yml']:
                import yaml
                x = yaml.safe_load(fo)

            # XML (.xml)
            elif ext == '.xml':
                import xml.etree.ElementTree as ET
                x = ET.parse(fo).getroot()

            # MAT (.mat)
            elif ext == '.mat':
                from scipy.io import loadmat
                x = loadmat(fo)

            # ZIP (.zip)
            elif ext == '.zip':
                import zipfile
                with zipfile.ZipFile(fo, 'r') as zip_ref:
                    x = {name: zip_ref.read(name) for name in zip_ref.namelist()}

            # MessagePack (.msgpack)
            elif ext == '.msgpack':
                import msgpack
                x = msgpack.unpackb(fo.read(), raw=False)

            # GeoJSON (.geojson)
            elif ext == '.geojson':
                import geopandas as gpd
                x = gpd.read_file(fo)

            # WAV (.wav)
            elif ext == '.wav':
                from scipy.io.wavfile import read as wav_read
                x = wav_read(fo)

            elif ext in ['.joblib']:
                import joblib
                x = joblib.load(fo, **kwargs)

            elif ext in ['.z', '.gz']:
                import gzip
                with gzip.open(fo, 'rb') as file:
                    x = file.read()
                    x = self.read_inner_content(x, **kwargs)

            elif ext in ['.bz2']:
                import bz2
                with bz2.open(fo, 'rb') as file:
                    x = file.read()
                    x = self.read_inner_content(x, **kwargs)

            elif ext in ['.xz', '.lzma']:
                import lzma
                with lzma.open(fo, 'rb') as file:
                    x = file.read()
                    x = self.read_inner_content(x, **kwargs)

            elif ext == '.safetensors':
                from safetensors.torch import load
                x = load(fo.read())

            else:
                x = fo.read()

        return x

    def read_inner_content(self, content, inner_ext=None, **kwargs):

        p = self.stem
        if '.' in p:
            inner_ext = inner_ext or f".{p.split('.')[-1]}"

        if inner_ext is not None:
            content = IOPath('/', data=content).read(ext=inner_ext, **kwargs)

        return content

    def readlines(self):
        return self.read(readlines=True, ext='.txt')

    def read_text(self):
        return self.read(ext='.txt')

    def read_bytes(self):
        return self.read(ext='.bin')

    @staticmethod
    def mode(op, ext):
        if op == 'write':
            m = 'w'
        else:
            m = 'r'

        if ext not in PureBeamPath.text_based_extensions:
            m = f"{m}b"

        return m

    def write_text(self, x, encoding=None, errors=None, newline=None):
        return self.write(x, ext='.txt')

    def write_bytes(self, x):
        return self.write(x, ext='.bin')

    def write(self, *args, ext=None, **kwargs):

        x = None
        if len(args) >= 1:
            x = args[0]

        if ext is None:
            ext = self.suffix

        path = str(self)

        with self(mode=PureBeamPath.mode('write', ext)) as fo:

            if ext == '.fea':

                if len(x.shape) == 1:
                    x = pd.Series(x)
                    if x.name is None:
                        x.name = 'val'

                x = pd.DataFrame(x)

                if isinstance(x.index, pd.MultiIndex):
                    raise TypeError("MultiIndex not supported with feather extension.")

                x = x.rename({c: str(c) for c in x.columns}, axis=1)

                index_name = x.index.name if x.index.name is not None else 'index'
                df = x.reset_index()
                new_name = PureBeamPath.feather_index_mark + index_name
                x = df.rename(columns={index_name: new_name})
                x.to_feather(fo, **kwargs)
            elif ext == '.csv':
                x = pd.DataFrame(x)
                x.to_csv(fo, **kwargs)
            elif ext == '.avro':
                import fastavro
                fastavro.writer(fo, x, **kwargs)
            elif ext in ['.pkl', '.pickle']:
                pd.to_pickle(x, fo, **kwargs)
            elif ext == '.npy':
                np.save(fo, x, **kwargs)
            elif ext == '.json':
                json.dump(x, fo, **kwargs)
            elif ext == '.ndjson':
                for xi in x:
                    json.dump(xi, fo, **kwargs)
                    fo.write("\n")
            elif ext == '.txt':
                fo.write(str(x))
            elif ext == '.npz':
                np.savez(fo, *args, **kwargs)
            elif ext == '.npzc':
                np.savez_compressed(fo, *args, **kwargs)
            elif ext in ['.adjlist', '.gexf', '.gml', '.pajek', '.graphml']:
                import networkx as nx
                write = getattr(nx, f'write_{ext[1:]}')
                write(x, fo, **kwargs)
            elif ext == '.scipy_npz':
                import scipy
                scipy.sparse.save_npz(fo, x, **kwargs)
                self.rename(f'{path}.npz', path)
            elif ext == '.parquet':
                x = pd.DataFrame(x)
                x.to_parquet(fo, **kwargs)
            elif ext == '.pt':
                import torch
                torch.save(x, fo, **kwargs)

            # HDF5 (.h5, .hdf5)
            elif ext in ['.h5', '.hdf5']:
                import h5py
                with h5py.File(fo, 'w') as f:
                    for key, value in x.items():
                        f.create_dataset(key, data=value)

            # YAML (.yaml, .yml)
            elif ext in ['.yaml', '.yml']:
                import yaml
                yaml.safe_dump(x, fo)

            # XML (.xml)
            elif ext == '.xml':
                import xml.etree.ElementTree as ET

                tree = ET.ElementTree(x)
                tree.write(fo)

            # MAT (.mat)
            elif ext == '.mat':
                from scipy.io import savemat
                savemat(fo, x)

            # ZIP (.zip)
            elif ext == '.zip':
                import zipfile

                with zipfile.ZipFile(fo, 'w') as zip_ref:
                    for name, content in x.items():
                        zip_ref.writestr(name, content)

            # MessagePack (.msgpack)
            elif ext == '.msgpack':
                import msgpack
                fo.write(msgpack.packb(x, use_bin_type=True))

            # GeoJSON (.geojson)
            elif ext == '.geojson':
                import geopandas as gpd
                gpd.GeoDataFrame(x).to_file(fo, driver='GeoJSON')

            # WAVt (.wav)
            elif ext == '.wav':
                from scipy.io.wavfile import write as wav_write
                wav_write(fo, *x)

            elif '.joblib' in ext:
                import joblib

                compress_methods = ext.split('_')[1:]
                if len(compress_methods) > 0:
                    cm = compress_methods[0]
                    compress_methods = {'z': 'zlib', 'gz': 'gzip', 'bz2': 'bz2', 'xz': 'xz', 'lzma': 'lzma'}
                    cm = compress_methods[cm]

                    if 'compress' not in kwargs:
                        kwargs['compress'] = cm
                    else:
                        kwargs['compress'] = (cm, kwargs['compress'])

                joblib.dump(x, fo, **kwargs)

            elif ext in ['.z', '.gz']:
                import gzip
                x = self.serialize_inner_content(x, **kwargs)
                with gzip.open(fo, 'wb') as file:
                    file.write(x)

            elif ext in ['.bz2']:
                import bz2
                x = self.serialize_inner_content(x, **kwargs)
                with bz2.open(fo, 'wb') as file:
                    file.write(x)

            elif ext in ['.xz', '.lzma']:
                import lzma
                x = self.serialize_inner_content(x, **kwargs)
                with lzma.open(fo, 'wb') as file:
                    file.write(x)

            elif ext == '.safetensors':
                from safetensors.torch import save
                raw_data = save(x, **kwargs)
                fo.write(raw_data)

            elif ext == '.ini':
                import configparser
                x = configparser.ConfigParser()
                x.read_dict(x)
                x.write(fo)

            elif ext in PureBeamPath.textual_extensions:
                assert isinstance(x, str), f"Expected str, got {type(x)}"
                fo.write(x)

            elif ext == '.bin':
                assert isinstance(x, bytes), f"Expected bytes, got {type(x)}"
                fo.write(x)

            else:
                raise ValueError(f"Unsupported extension type: {ext} for file {x}.\n"
                                 f"Use write_bytes or write_text instead,  \n"
                                 f"or use one of the existing extensions by explicitly setting the ext argument,\n"
                                 f"e.g. path.write(content, ext=.pkl).")

        return self

    def serialize_inner_content(self, content, inner_ext=None, **kwargs):

        p = self.stem
        if '.' in p:
            inner_ext = inner_ext or f".{p.split('.')[-1]}"
        inner_ext = inner_ext or '.pkl'
        io_path = IOPath('/').write(content, ext=inner_ext, **kwargs)
        return io_path.data['/']

    def resolve(self, strict=False):
        return self.gen(self.path.resolve(strict=strict))


class DictBasedPath(PureBeamPath):
    def __init__(self, *pathsegments, scheme=None, client=None, data=None, **kwargs):
        super().__init__(*pathsegments, scheme=scheme, **kwargs)
        if client is None:
            client = {}
        if data is not None:
            client = data
        self.client = client

    @property
    def data(self):
        return self.get_data()

    def get_data(self):
        return self.client

    def set_data(self, data):
        self.client = data

    def is_file(self):
        raise NotImplementedError

    def is_dir(self):
        client = self.client
        if len(self.parts) == 1:
            return True
        for p in self.parts[1:]:
            if not isinstance(client, dict):
                return False
            if p not in client:
                return False
            client = client[p]
        return isinstance(client, dict)

    def exists(self):
        client = self.client
        if len(self.parts) == 1:
            return True
        for p in self.parts[1:]:
            if p not in client:
                return False
        return True

    @property
    def _obj(self):
        client = self.client
        for p in self.parts[1:]:
            client = client[p]
        return client

    @property
    def _parent(self):
        client = self.client
        for p in self.parts[1:-1]:
            client = client[p]
        return client

    def iterdir(self):
        client = self._obj
        for p in client:
            yield self.gen(self.path.joinpath(p))

    def mkdir(self, *args, parents=True, exist_ok=True, **kwargs):

        if self.is_root():
            return

        if not exist_ok:
            if self.exists():
                raise FileExistsError

        if not parents:
            if self._parent is None:
                raise FileNotFoundError

        client = self.client
        for p in self.parts[1:]:
            if p not in client:
                client[p] = {}
            elif p in client and not isinstance(client[p], dict):
                raise NotADirectoryError

    def rmdir(self):
        client = self._parent
        del client[self.parts[-1]]

    def unlink(self, missing_ok=False):

        if self.is_root():
            raise ValueError("Cannot delete root")

        client = self._parent
        try:
            del client[self.parts[-1]]
        except KeyError as e:
            if not missing_ok:
                raise e

    def rename(self, target):

        if target.is_root():
            raise ValueError("Cannot rename to root")

        if type(target) is str:
            target = self.gen(target)
        target_parent = target._parent
        target_parent[target.parts[-1]] = self._obj
        self.unlink()

    def replace(self, target):
        self.rename(target)

    def __enter__(self):
        raise NotImplementedError

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError


class IOPath(DictBasedPath):

    # this class represents a in memory file system in the form of dictionaries of dictionaries of bytes/strings objects
    # like other PureBeamPath implementations, it is a pathlib/beam_path api

    def __init__(self, *pathsegments, client=None, data=None, **kwargs):
        super().__init__(*pathsegments, scheme='io', client=client, data=data, **kwargs)

    def is_file(self):
        client = self.client
        if len(self.parts) == 1:
            return False
        for p in self.parts[1:]:
            if not isinstance(client, dict):
                return False
            if p not in client:
                return False
            client = client[p]
        return isinstance(client, (bytes, str))

    def __enter__(self):
        if self.mode in ["rb", "r"]:
            self.file_object = BytesIO(self._obj)
        elif self.mode in ['wb', 'w']:
            self.file_object = BytesIO() if 'b' in self.mode else StringIO(newline=self.open_kwargs['newline'])
        else:
            raise ValueError

        return self.file_object

    def __exit__(self, exc_type, exc_val, exc_tb):

        if self.mode in ["wb", "w"]:
            parent = self._parent
            name = self.parts[-1]
            self.file_object.seek(0)
            content = self.file_object.getvalue()
            parent[name] = content

        self.close_at_exit()


class DictPath(DictBasedPath):

    def __init__(self, *pathsegments, client=None, data=None, **kwargs):
        super().__init__(*pathsegments, scheme='dict', client=client, data=data, **kwargs)

    def is_file(self):
        client = self.client
        if len(self.parts) == 1:
            return False
        for p in self.parts[1:]:
            if not isinstance(client, dict):
                return False
            if p not in client:
                return False
            client = client[p]
        return isinstance(client, BeamFile)

    def write(self, *args, ext=None, **kwargs):
        assert len(args) == 1, "DictPath.write takes exactly one argument"
        x = args[0]
        self._parent[self.parts[-1]] = BeamFile(x, timestamp=datetime.now())

    def read(self, ext=None, **kwargs):
        return self._obj.data


class BeamKey:

    def __init__(self, config_path=None, **kwargs):
        self.keys = {}

        self._config_path = config_path
        if self._config_path is None:
            self._config_path = Path.home().joinpath('conf.pkl')

        self._config_file = None
        self.hparams = kwargs

    def set_hparams(self, hparams):

        for k, v in hparams.items():
            self.hparams[k] = v
        # clear config file
        self._config_path = None

    @property
    def config_path(self):
        if self._config_path is None:
            if 'config_file' in self.hparams:
                self._config_path = Path(self.hparams['config_file'])
            else:
                self._config_path = Path.home().joinpath('conf.pkl')
        return self._config_path

    @property
    def config_file(self):
        if self._config_file is None:
            if self.config_path is not None and self.config_path.is_file():
                self._config_file = pd.read_pickle(self.config_path)
        return self._config_file

    def store(self, name=None, value=None, store_to_env=True):
        if name is not None:
            self.keys[name] = value

        # store to environment
        if store_to_env:
            os.environ[name] = str(value)

        config_file = self.config_file
        if config_file is None:
            config_file = {}

        for k, v in self.keys.items():
            config_file[k] = v

        self.config_path.parent.mkdir(parents=True, exist_ok=True)
        pd.to_pickle(config_file, self.config_path)
        self._config_file = config_file

    def __setitem__(self, key, value):
        self(key, value)

    def __getitem__(self, item):
        return self(item)

    def __call__(self, name, value=None, store=True):

        if value is not None:
            self.keys[name] = value
            if store:
                self.store(name, value)
            return value
        elif name in self.keys:
            value = self.keys[name]
        elif name in self.hparams and self.hparams[name] is not None:
            value = self.hparams[name]
            self.keys[name] = value
        elif name in os.environ:
            value = os.environ[name]
            self.keys[name] = value
        elif self.config_file is not None and name in self.config_file:
            value = self.config_file[name]
            self.keys[name] = value
        else:
            ValueError(f"Cannot find key: {name} in BeamKey")

        return value



import os
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ['TOKENIZERS_PARALLELISM'] = 'true'

from .utils import tqdm_beam as tqdm

__all__ = ['UniversalBatchSampler', 'UniversalDataset',
           'Experiment', 'beam_algorithm_generator',
           'NeuralAlgorithm',
           'LinearNet', 'PackedSet', 'copy_network', 'reset_network', 'DataTensor', 'BeamOptimizer', 'BeamScheduler',
           'BeamNN',
           'BeamData',
           'slice_to_index', 'beam_device', 'as_tensor', 'batch_augmentation', 'as_numpy', 'DataBatch', 'beam_hash',
           'UniversalConfig', 'beam_arguments', 'BeamConfig', 'BeamParam',
           'check_type', 'Timer',
           'beam_logger', 'beam_kpi',
           'beam_path', 'beam_key',
           'beam_server', 'beam_client',
           '__version__',
           'resource',
           'KeysConfig',
           'tqdm',
           ]


try:
    import torch
    has_torch = True
except ImportError:
    has_torch = False

if has_torch:
    from .dataset import UniversalBatchSampler, UniversalDataset
    from .experiment import Experiment, beam_algorithm_generator
    from .algorithm import NeuralAlgorithm
    from .nn import LinearNet, PackedSet, copy_network, reset_network, DataTensor, BeamOptimizer, BeamScheduler, BeamNN
    from .data import BeamData
    from .utils import slice_to_index, beam_device, as_tensor, batch_augmentation, as_numpy, DataBatch, beam_hash

from .config import UniversalConfig, beam_arguments, BeamConfig, BeamParam
from .utils import check_type, Timer
from .logger import beam_logger, beam_kpi

from functools import partial
Timer = partial(Timer, logger=beam_logger)

from .path import beam_path, beam_key
from .serve import beam_server, beam_client
from ._version import __version__
from .resource import resource

from .config import KeysConfig
beam_key.set_hparams(KeysConfig(silent=True, strict=True))





def resource(uri, **kwargs):
    if type(uri) != str:
        return uri
    if ':' not in uri:
        from .path import beam_path
        return beam_path(uri, **kwargs)

    scheme = uri.split(':')[0]
    if scheme in ['file', 's3', 's3-pa', 'hdfs', 'hdfs-pa', 'sftp', 'comet', 'io', 'dict', 'redis', 'smb', 'nt']:
        from .path import beam_path
        return beam_path(uri, **kwargs)
    elif scheme in ['beam-http', 'beam-https', 'beam-grpc']:
        from .serve import beam_client
        return beam_client(uri, **kwargs)
    elif scheme in ['async-http', 'async-https']:
        from .distributed import async_client
        return async_client(uri, **kwargs)
    elif scheme in ['openai', 'tgi', 'fastchat', 'huggingface', 'fastapi', 'fastapi-dp']:
        from .llm import beam_llm
        return beam_llm(uri, **kwargs)
    elif scheme in ['triton', 'triton-http', 'triton-grpc', 'triton-https', 'triton-grpcs']:
        from .serve import triton_client
        return triton_client(uri, **kwargs)
    else:
        raise Exception(f'Unknown resource scheme: {scheme}')



import copy

import ray
from ray.air import RunConfig
from ray.tune.schedulers import ASHAScheduler
from ray.tune.search.optuna import OptunaSearch

from .utils import TimeoutStopper
from ..utils import find_port, check_type, is_notebook, beam_device
from ..logger import beam_logger as logger
from ..path import beam_path, BeamPath
from ..distributed.ray_dispatcher import RayCluster

from ray.tune import JupyterNotebookReporter, TuneConfig
from ray import tune, train
from functools import partial
from ..experiment import Experiment

import numpy as np
from .core import BeamHPO


class RayHPO(BeamHPO, RayCluster):

    @staticmethod
    def _categorical(param, choices):
        return tune.choice(choices)

    @staticmethod
    def _uniform(param, start, end):
        return tune.uniform(start, end)

    @staticmethod
    def _loguniform(param, start, end):
        return tune.loguniform(start, end)

    @staticmethod
    def _linspace(param, start, end, n_steps, endpoint=True, dtype=None):
        x = np.linspace(start, end, n_steps, endpoint=endpoint)
        step_size = (end - start) / n_steps
        end = end - step_size * (1 - endpoint)

        if np.sum(np.abs(x - np.round(x))) < 1e-8 or dtype in [int, np.int, np.int64, 'int', 'int64']:

            start = int(np.round(start))
            step_size = int(np.round(step_size))
            end = int(np.round(end))

            return tune.qrandint(start, end, step_size)

        return tune.quniform(start, end, (end - start) / n_steps)

    @staticmethod
    def _logspace(param, start, end, n_steps, base=None, dtype=None):

        if base is None:
            base = 10

        emin = base ** start
        emax = base ** end

        x = np.logspace(start, end, n_steps, base=base)

        if np.sum(np.abs(x - np.round(x))) < 1e-8 or dtype in [int, np.int, np.int64, 'int', 'int64']:
            base = int(x[1] / x[0])
            return tune.lograndint(int(emin), int(emax), base=base)

        step_size = (x[1] / x[0]) ** ( (end - start) / n_steps )
        return tune.qloguniform(emin, emax, step_size, base=base)

    @staticmethod
    def _randn(param, mu, sigma):
        return tune.qrandn(mu, sigma)

    def runner(self, config):

        hparams = self.generate_hparams(config)

        experiment = Experiment(hparams, hpo='tune', print_hyperparameters=False)

        trial_dir = beam_path(train.get_context().get_trial_dir())
        logger.info(f"Experiment directory is: {experiment.experiment_dir}, see experiment_dir at beam_configuration.pkl"
                    f" in trial directory ({trial_dir})")
        trial_dir.joinpath('beam_configuration.pkl').write({'experiment_dir': experiment.experiment_dir})

        alg, report = experiment.fit(alg=self.alg, algorithm_generator=self.ag, return_results=True)
        train.report({report.objective_name: report.best_objective})
        if self.post_train_hook is not None:
            self.post_train_hook(alg=alg, experiment=experiment, hparams=hparams, suggestion=config, results=report)

        self.tracker(algorithm=alg, results=report.data, hparams=hparams, suggestion=config)

    def run(self, *args, runtime_env=None, tune_config_kwargs=None, run_config_kwargs=None,
            init_config_kwargs=None, restore_path=None, restore_config=None, **kwargs):

        hparams = copy.deepcopy(self.hparams)
        hparams.update(kwargs)

        search_space = self.get_suggestions()

        # the ray init configuation

        ray_address = self.hparams.get('ray_address')
        init_config_kwargs = init_config_kwargs or {}

        if ray_address != 'auto':

            dashboard_port = find_port(port=self.hparams.get('dashboard_port'),
                                       get_port_from_beam_port_range=self.hparams.get('get_port_from_beam_port_range'))
            logger.info(f"Opening ray-dashboard on port: {dashboard_port}")
            include_dashboard = self.hparams.get('include_dashboard')

        else:

            dashboard_port = None
            include_dashboard = False

        self.init_ray(address=ray_address, include_dashboard=include_dashboard, dashboard_port=dashboard_port,
                      runtime_env=runtime_env, **init_config_kwargs)

        # the ray tune configuation
        stop = kwargs.get('stop', None)
        train_timeout = hparams.get('train-timeout')
        if train_timeout is not None and train_timeout > 0:
            stop = TimeoutStopper(train_timeout)

        # fix gpu to device 0
        if self.experiment_hparams.get('device') != 'cpu':
            self.experiment_hparams.set('device', 'cuda')

        runner_tune = tune.with_resources(
                tune.with_parameters(partial(self.runner)),
                resources={"cpu": hparams.get('cpus-per-trial'),
                           "gpu": hparams.get('gpus-per-trial')}
            )

        tune_config_kwargs = tune_config_kwargs or {}
        if 'metric' not in tune_config_kwargs.keys():
            tune_config_kwargs['metric'] = self.experiment_hparams.get('objective')
        if 'mode' not in tune_config_kwargs.keys():
            mode = self.experiment_hparams.get('objective-mode')
            tune_config_kwargs['mode'] = self.get_optimization_mode(mode, tune_config_kwargs['metric'])

        if 'progress_reporter' not in tune_config_kwargs.keys() and is_notebook():
            tune_config_kwargs['progress_reporter'] = JupyterNotebookReporter(overwrite=True)

        tune_config_kwargs['num_samples'] = self.hparams.get('n_trials')
        tune_config_kwargs['max_concurrent_trials'] = self.hparams.get('n_jobs', 1)

        # if 'scheduler' not in tune_config_kwargs.keys():
        #     tune_config_kwargs['scheduler'] = ASHAScheduler()

        if 'search_alg' not in tune_config_kwargs.keys():
            metric = tune_config_kwargs['metric']
            mode = tune_config_kwargs['mode']
            tune_config_kwargs['search_alg'] = OptunaSearch(space=None, metric=metric, mode=mode)
            # tune_config_kwargs['search_alg'] = OptunaSearch()

        tune_config = TuneConfig(**tune_config_kwargs)

        # the ray run configuration
        local_dir = self.hparams.get('hpo_path')

        run_config_kwargs = run_config_kwargs or {}
        if 'name' not in run_config_kwargs.keys():
            run_config_kwargs['name'] = self.identifier
        if 'stop' not in run_config_kwargs.keys():
            run_config_kwargs['stop'] = stop
        if 'storage_path' not in run_config_kwargs.keys():
            run_config_kwargs['storage_path'] = local_dir
        run_config = RunConfig(**run_config_kwargs)

        logger.info(f"Starting ray-tune hyperparameter optimization process. "
                    f"Results and logs will be stored at {local_dir}")

        if restore_path:
            restore_config = restore_config or {}
            tuner = tune.Tuner.restore(restore_path, runner_tune, **restore_config)
        else:
            tuner = tune.Tuner(runner_tune, param_space=search_space, tune_config=tune_config, run_config=run_config)
        analysis = tuner.fit()

        return analysis



from ..algorithm import NeuralAlgorithm
from ..path import local_copy, in_memory_storage
from transformers import AutoModelForCausalLM, AutoTokenizer, AutoConfig


class FineTuneLLM(NeuralAlgorithm):

    def __init__(self, hparams, **kwargs):

        networks = None
        config = None
        if hparams.get('reload_path', None) is None:
            config = self.load_configuration(hparams)
            model = self.load_lora_model(hparams, config=config)
            networks = {'llm': model}

        # dtype = hparams.get('model_dtype', 'float16')
        # if dtype != 'float32':
        #     dtype = getattr(torch, dtype)
        #     for n, v in networks['llm'].named_parameters():
        #         if not v.requires_grad:
        #             v.to(dtype=dtype)

        self.tokenizer = self.load_tokenizer(hparams, config=config)
        super().__init__(hparams, networks=networks, **kwargs)

    @staticmethod
    def load_configuration(hparams):
        return AutoConfig.from_pretrained(hparams.model, cache_dir=hparams.hf_cache_dir)

    @staticmethod
    def load_tokenizer(hparams, config=None):
        if config is None:
            config = FineTuneLLM.load_configuration(hparams)
        return AutoTokenizer.from_pretrained(hparams.model, config=config)

    @staticmethod
    def load_model(hparams, config=None):

        if config is None:
            config = FineTuneLLM.load_configuration(hparams)

        model = AutoModelForCausalLM.from_pretrained(hparams.model, config=config, load_in_8bit=hparams.load_in_8bit)
        return model

    @staticmethod
    def load_lora_model(hparams, config=None, model=None):

        if model is None:
            model = FineTuneLLM.load_model(hparams, config=config)

        from peft import LoraConfig, get_peft_model
        lora_config = LoraConfig(r=hparams.lora_r, lora_alpha=hparams.lora_alpha,
                                 target_modules=hparams.target_modules, lora_dropout=hparams.lora_dropout,
                                 bias=hparams.lora_bias, fan_in_fan_out=hparams.lora_fan_in_fan_out,
                                 modules_to_save=hparams.modules_to_save,
                                 layers_to_transform=hparams.layers_to_transform,
                                 task_type="CAUSAL_LM")

        model = get_peft_model(model, lora_config)
        return model

    def train_iteration(self, sample=None, label=None, index=None, counter=None, subset=None, training=True, **kwargs):
        net = self.networks['llm']

        # labels are shifted inside the model forward pass
        # (see https://github.com/huggingface/transformers/blob/main/src/transformers/models/llama/modeling_llama.py#L1106)

        labels = sample['labels'] if 'labels' in sample else sample['input_ids']

        res = net(input_ids=sample['input_ids'], attention_mask=sample['attention_mask'], labels=labels)
        self.apply(res.loss)

    def save_checkpoint(self, path=None, networks=True, optimizers=True, schedulers=True,
                        processors=True, scaler=True, scalers=True, swa_schedulers=True, swa_networks=True,
                        hparams=True, aux=None, pickle_model=False):

        in_memory_path = in_memory_storage(mode='file')
        with local_copy(in_memory_path, as_beam_path=False) as local_path:
            self.networks['llm'].save_pretrained(local_path)

        if networks:
            if aux is None:
                aux = {}
            aux['lora'] = in_memory_path.data

        return super().save_checkpoint(path=path, networks=False, optimizers=optimizers, schedulers=schedulers,
                        processors=processors, scaler=scaler, scalers=scalers, swa_schedulers=swa_schedulers, swa_networks=swa_networks,
                        hparams=hparams, aux=aux, pickle_model=pickle_model)

    def load_checkpoint(self, path_or_state, strict=True, networks=True, optimizers=True, schedulers=True,
                        processors=True, scaler=True, scalers=True, swa_schedulers=True, swa_networks=True,
                        hparams=True, load_epoch=True):

        aux = super().load_checkpoint(path_or_state, networks=False, optimizers=optimizers, schedulers=schedulers,
                        processors=processors, scaler=scaler, scalers=scalers, swa_schedulers=swa_schedulers, swa_networks=swa_networks,
                        hparams=hparams)

        from peft import PeftModel

        in_memory_path = in_memory_storage(mode='file', data=aux.pop('lora'))
        with local_copy(in_memory_path, as_beam_path=False) as local_path:
            model = self.load_model(self.hparams)
            model = PeftModel.from_pretrained(model, local_path, is_trainable=True)
            self.add_network(model, 'llm')
            # self.networks['llm'].from_pretrained(local_path)

        return aux


import itertools
import os, sys
from collections import defaultdict
import random
import numpy as np
from fnmatch import filter
from tqdm.notebook import tqdm as tqdm_notebook
from tqdm import tqdm

import pandas as pd
import __main__
from datetime import timedelta
import time
import re
import threading

import traceback
import linecache
from concurrent.futures import ThreadPoolExecutor, TimeoutError


try:
    import modin.pandas as mpd

    has_modin = True
except ImportError:
    mpd = None
    has_modin = False

try:
    import torch

    has_torch = True
except ImportError:
    has_torch = False

try:
    import scipy

    has_scipy = True
except ImportError:
    has_scipy = False

try:
    import polars as pl
    has_polars = True
except ImportError:
    has_polars = False

# try:
#     import cudf
#     has_cudf = True
# except ImportError:
#     has_cudf = False

import socket
from contextlib import closing
from collections import namedtuple
from timeit import default_timer as timer

from pathlib import Path, PurePath, PurePosixPath
from collections import Counter
import inspect
from argparse import Namespace
from functools import wraps, partial
from ..path import PureBeamPath


TypeTuple = namedtuple('TypeTuple', 'major minor element')
DataBatch = namedtuple("DataBatch", "index label data")


# class Beamdantic(BaseModel):
#     # To be used with pydantic classes and lazy_property
#     _lazy_cache: Any = PrivateAttr(default=None)


class BeamDict(dict, Namespace):
    def __init__(self, initial_data=None, **kwargs):
        if isinstance(initial_data, dict):
            self.__dict__.update(initial_data)
        elif isinstance(initial_data, BeamDict):
            self.__dict__.update(initial_data.__dict__)
        elif hasattr(initial_data, '__dict__'):  # This will check for Namespace or any other object with attributes
            self.__dict__.update(initial_data.__dict__)
        elif initial_data is not None:
            raise TypeError(
                "initial_data should be either a dictionary, an instance of DictNamespace, or a Namespace object")

            # Handle additional kwargs
        for key, value in kwargs.items():
            self.__dict__[key] = value

        for key, value in self.__dict__.items():
            if isinstance(value, dict):
                self.__dict__[key] = BeamDict(value)
            if isinstance(value, list):
                self.__dict__[key] = [BeamDict(v) if isinstance(v, dict) else v for v in value]

    def __getattr__(self, key):
        try:
            return self.__dict__[key]
        except KeyError:
            raise AttributeError(f"'{type(self).__name__}' object has no attribute '{key}'")

    def pop(self, key, default=None):
        try:
            return self.__dict__.pop(key)
        except KeyError:
            return default

    def items(self):
        return self.__dict__.items()

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def __setattr__(self, key, value):
        self.__dict__[key] = value

    def __getitem__(self, key):
        return self.__dict__[key]

    def __setitem__(self, key, value):
        self.__dict__[key] = value

    def __repr__(self):
        return repr(self.__dict__)

    def __contains__(self, key):
        return key in self.__dict__


def retrieve_name(var):
    for fi in reversed(inspect.stack()):
        names = [var_name for var_name, var_val in fi.frame.f_locals.items() if var_val is var]
        if len(names) > 0:
            return names[0]

# def retrieve_name(var):
#     name = None
#     # Start by checking the global scope of the caller.
#     for fi in reversed(inspect.stack()):
#         names = [var_name for var_name, var_val in fi.frame.f_globals.items() if var_val is var]
#         if names:
#             name = names[0]
#             break  # Exit on the first match in global scope.
#
#     # If not found in global scope, check the local scope from inner to outer.
#     if not name:
#         for fi in inspect.stack():
#             names = [var_name for var_name, var_val in fi.frame.f_locals.items() if var_val is var]
#             if names:
#                 name = names[0]
#                 break  # Exit on the first match in local scope.
#
#     return name


def has_kwargs(func):
    return any(p.kind == inspect.Parameter.VAR_KEYWORD for p in inspect.signature(func).parameters.values())


def strip_prefix(text, prefix):
    if text.startswith(prefix):
        return text[len(prefix):]
    return text


def strip_suffix(text, suffix):
    if text.endswith(suffix):
        return text[:-len(suffix)]
    return text


class nested_defaultdict(defaultdict):

    @staticmethod
    def default_factory_list():
        return defaultdict(list)

    @staticmethod
    def default_factory_dict():
        return defaultdict(dict)

    def __init__(self, default_factory=None, **kwargs):
        if default_factory is list:
            default_factory = self.default_factory_list
        elif default_factory is dict:
            default_factory = self.default_factory_dict
        super().__init__(default_factory, **kwargs)


def lazy_property(fn):

    @property
    def _lazy_property(self):
        try:
            cache = getattr(self, '_lazy_cache')
            return cache[fn.__name__]
        except KeyError:
            v = fn(self)
            cache[fn.__name__] = v
            return v
        except AttributeError:
            v = fn(self)
            setattr(self, '_lazy_cache', {fn.__name__: v})
            return v

    @_lazy_property.setter
    def _lazy_property(self, value):
        try:
            cache = getattr(self, '_lazy_cache')
            cache[fn.__name__] = value
        except AttributeError:
            setattr(self, '_lazy_cache', {fn.__name__: value})

    return _lazy_property


def get_public_ip():
    import requests
    try:
        response = requests.get("https://api64.ipify.org?format=json")
        ip = response.json().get("ip")
        return ip
    except requests.RequestException:
        return None


def rate_string_format(n, t):
    if n / t > 1:
        return f"{n / t: .4} [iter/sec]"
    return f"{t / n: .4} [sec/iter]"


def beam_base_port():
    base_range = None
    if 'JUPYTER_PORT' in os.environ:
        base_range = 100 * (int(os.environ['JUPYTER_PORT']) // 100)
    elif os.path.isfile('/workspace/configuration/config.csv'):
        try:
            conf = pd.read_csv('/workspace/configuration/config.csv', index_col=0)
            base_range = 100 * int(conf.loc['initials'])
        except:
            pass
    return base_range


def beam_service_port(service):
    port = None
    try:
        conf = pd.read_csv('/workspace/configuration/config.csv', index_col=0)
        port = int(conf.drop_duplicates().loc[service.lower()])
    except:
        pass

    return port


def find_port(port=None, get_port_from_beam_port_range=True, application='none', blacklist=None, whitelist=None):
    from ..logger import beam_logger as logger

    if blacklist is None:
        blacklist = []

    if whitelist is None:
        whitelist = []

    blacklist = [int(p) for p in blacklist]
    whitelist = [int(p) for p in whitelist]

    if application == 'tensorboard':
        first_beam_range = 66
        first_global_range = 26006
    elif application == 'flask':
        first_beam_range = 50
        first_global_range = 25000
    elif application == 'ray':
        first_beam_range = 65
        first_global_range = 28265
    elif application == 'distributed':
        first_beam_range = 64
        first_global_range = 28264
    else:
        first_beam_range = 2
        first_global_range = 30000

    if port is None:

        port_range = None

        if get_port_from_beam_port_range:

            base_range = None
            if os.path.isfile('/workspace/configuration/config.csv'):
                conf = pd.read_csv('/workspace/configuration/config.csv')
                base_range = int(conf.set_index('parameters').drop_duplicates().loc['initials'])

            if base_range is not None:
                port_range = range(base_range * 100, (base_range + 1) * 100)
                port_range = np.roll(np.array(port_range), -first_beam_range)

        if port_range is None:
            port_range = np.roll(np.array(range(10000, 2 ** 16)), -first_global_range)

        for p in port_range:

            if p in blacklist:
                continue

            if whitelist and p not in whitelist:
                continue

            if check_if_port_is_available(p):
                port = str(p)
                break

        if port is None:
            logger.error("Cannot find free port in the specified range")
            return

    else:
        if not check_if_port_is_available(port):
            logger.error(f"Port {port} is not available")
            return

    return port


def is_boolean(x):
    x_type = check_type(x)
    if x_type.minor in ['numpy', 'pandas', 'tensor'] and 'bool' in str(x.dtype).lower():
        return True
    if x_type.minor == 'list' and len(x) and isinstance(x[0], bool):
        return True

    return False


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        p = str(s.getsockname()[1])
    return p


def check_if_port_is_available(port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    return sock.connect_ex(('127.0.0.1', int(port))) != 0


def get_notebook_name():
    """Execute JS code to save Jupyter notebook name to variable `notebook_name`"""
    from IPython.core.display import Javascript, display_javascript
    js = Javascript("""IPython.notebook.kernel.execute('notebook_name = "' + IPython.notebook.notebook_name + '"');""")

    return display_javascript(js)


def pretty_format_number(x, short=False):

    just = 4 if short else 10
    trim = 4 if short else 8
    exp = 2 if short else 4

    if x is None or np.isinf(x) or np.isnan(x):
        return f'{x}'.ljust(just)
    if int(x) == x and np.abs(x) < 10000:
        return f'{int(x)}'.ljust(just)
    if np.abs(x) >= 10000 or np.abs(x) < 0.0001:
        return f'{float(x):.4}'.ljust(just)
    if np.abs(x) >= 1000:
        return f'{x:.1f}'.ljust(just)
    if np.abs(x) < 10000 and np.abs(x) >= 0.0001:
        nl = int(np.log10(np.abs(x)))
        return f'{np.sign(x) * int(np.abs(x) * (10 ** (exp - nl))) * float(10 ** (nl - exp))}'.ljust(trim)[:trim].ljust(just)

    return f'{x}:NoFormat'


def pretty_print_timedelta(seconds):
    # Convert seconds into timedelta
    t_delta = timedelta(seconds=seconds)

    # Extract days, hours, minutes and seconds
    days = t_delta.days
    if days > 0:
        seconds = t_delta.seconds
        frac_days = days + seconds / (3600 * 24)
        return f"{pretty_format_number(frac_days, short=True)} days"

    hours = t_delta.seconds // 3600
    if hours > 0:
        seconds = t_delta.seconds % 3600
        frac_hours = hours + seconds / 3600
        return f"{pretty_format_number(frac_hours, short=True)} hours"

    minutes = t_delta.seconds // 60
    if minutes > 0:
        seconds = t_delta.seconds % 60
        frac_minutes = minutes + seconds / 60
        return f"{pretty_format_number(frac_minutes, short=True)} minutes"

    return f"{pretty_format_number(t_delta.seconds, short=True)} seconds"


def check_element_type(x):
    unknown = (check_minor_type(x) == 'other')

    if not unknown and not np.isscalar(x) and (has_torch and not (torch.is_tensor(x) and (not len(x.shape)))):
        return 'array'

    if pd.isna(x):
        return 'none'

    if hasattr(x, 'dtype'):
        # this case happens in custom classes that have a dtype attribute
        if unknown:
            return 'other'

        t = str(x.dtype).lower()
    else:
        t = str(type(x)).lower()

    if 'int' in t:
        return 'int'
    if 'bool' in t:
        return 'bool'
    if 'float' in t:
        return 'float'
    if 'str' in t:
        return 'str'
    if 'complex' in t:
        return 'complex'

    return 'object'


def check_minor_type(x):
    if has_torch and isinstance(x, torch.Tensor):
        return 'tensor'
    if isinstance(x, np.ndarray):
        return 'numpy'
    if isinstance(x, pd.core.base.PandasObject):
        return 'pandas'
    if isinstance(x, dict):
        return 'dict'
    if isinstance(x, list):
        return 'list'
    if isinstance(x, tuple):
        return 'tuple'
    if isinstance(x, set):
        return 'set'
    if has_modin and isinstance(x, mpd.base.BasePandasDataset):
        return 'modin'
    if has_scipy and scipy.sparse.issparse(x):
        return 'scipy_sparse'
    if has_polars and isinstance(x, pl.DataFrame):
        return 'polars'
    if isinstance(x, PurePath) or isinstance(x, PureBeamPath):
        return 'path'
    else:
        return 'other'

def elt_of_list(x):
    if len(x) < 100:
        sampled_indices = range(len(x))
    else:
        sampled_indices = np.random.randint(len(x), size=(100,))

    elt0 = None
    for i in sampled_indices:
        elt = check_element_type(x[i])

        if elt0 is None:
            elt0 = elt

        if elt != elt0:
            return 'object'

    return elt0


def check_type(x, check_minor=True, check_element=True):
    '''

    returns:

    <major type>, <minor type>, <elements type>

    major type: container, array, scalar, none, other
    minor type: dict, list, tuple, set, tensor, numpy, pandas, scipy_sparse, native, none
    elements type: array, int, float, complex, bool, str, object, empty, none, unknown

    '''

    if np.isscalar(x) or (has_torch and torch.is_tensor(x) and (not len(x.shape))):
        mjt = 'scalar'
        if check_minor:
            if type(x) in [int, float, str, complex, bool]:
                mit = 'native'
            else:
                mit = check_minor_type(x)
        else:
            mit = 'na'
        elt = check_element_type(x) if check_element else 'na'

    elif isinstance(x, dict):
        mjt = 'container'
        mit = 'dict'

        if check_element:
            if len(x):
                elt = check_element_type(next(iter(x.values())))
            else:
                elt = 'empty'
        else:
            elt = 'na'

    elif x is None:
        mjt = 'none'
        mit = 'none'
        elt = 'none'

    elif isinstance(x, slice):
        mjt = 'slice'
        mit = 'slice'
        elt = 'slice'

    elif isinstance(x, Counter):
        mjt = 'counter'
        mit = 'counter'
        elt = 'counter'

    else:

        elt = 'unknown'

        if hasattr(x, '__len__'):
            mjt = 'array'
        else:
            mjt = 'other'
        if isinstance(x, list) or isinstance(x, tuple) or isinstance(x, set):
            if not len(x):
                elt = 'empty'
            else:

                if len(x) < 20:
                    elts = [check_element_type(xi) for xi in x]

                else:

                    sample_size = 20
                    try:
                        ind = np.random.randint(len(x), size=(sample_size,))
                        elts = [check_element_type(x[i]) for i in ind]
                    except TypeError:
                        # assuming we are in the case of a set
                        random.sample(list(x), sample_size)

                set_elts = set(elts)
                if len(set_elts) == 1:
                    elt = elts[0]
                elif set_elts == {'int', 'float'}:
                    elt = 'float'
                else:
                    elt = 'object'

            if elt in ['array', 'object', 'none']:
                mjt = 'container'

        mit = check_minor_type(x) if check_minor else 'na'

        if elt:
            if mit in ['numpy', 'tensor', 'pandas', 'scipy_sparse']:
                if mit == 'pandas':
                    dt = str(x.values.dtype)
                else:
                    dt = str(x.dtype)
                if 'float' in dt:
                    elt = 'float'
                elif 'int' in dt:
                    elt = 'int'
                else:
                    elt = 'object'

        if mit == 'other':
            mjt = 'other'
            elt = 'other'

    return TypeTuple(major=mjt, minor=mit, element=elt)


def include_patterns(*patterns):
    """Factory function that can be used with copytree() ignore parameter.
    Arguments define a sequence of glob-style patterns
    that are used to specify what files to NOT ignore.
    Creates and returns a function that determines this for each directory
    in the file hierarchy rooted at the source directory when used with
    shutil.copytree().
    """

    def _ignore_patterns(path, names):
        keep = set(name for pattern in patterns
                   for name in filter(names, pattern))
        ignore = set(name for name in names
                     if name not in keep and not os.path.isdir(os.path.join(path, name)))
        return ignore

    return _ignore_patterns


def running_platform() -> str:
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return 'notebook'  # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return 'ipython'  # Terminal running IPython
        else:
            return 'other'  # Other type (?)
    except NameError:
        if hasattr(__main__, '__file__'):
            return 'script'
        else:
            return 'console'


def is_notebook() -> bool:
    return running_platform() == 'notebook'


def recursive_func(x, func, *args, **kwargs):
    if isinstance(x, dict):
        return {k: recursive_func(v, func, *args, **kwargs) for k, v in x.items()}
    elif isinstance(x, list):
        return [recursive_func(s, func, *args, **kwargs) for s in x]
    elif x is None:
        return None
    else:
        return func(x, *args, **kwargs)


def squeeze_scalar(x, x_type=None):

    if x_type is None:
        x_type = check_type(x)

    if x_type.minor == 'list':
        if len(x) == 1:
            x = x[0]
            x_type = check_type(x)

    if x_type.major == 'scalar':
        if x_type.element == 'int':
            return int(x)
        elif x_type.element == 'float':
            return float(x)
        elif x_type.element == 'complex':
            return complex(x)
        elif x_type.element == 'bool':
            return bool(x)
        elif x_type.element == 'str':
            return str(x)

    return x


def dictionary_iterator(d):

    d = {k: iter(v) for k, v in d.items()}
    for _ in itertools.count():
        try:
            yield {k: next(v) for k, v in d.items()}
        except StopIteration:
            return


def get_item_with_tuple_key(x, key):
    if x is None:
        return None

    if isinstance(key, tuple):
        for k in key:
            if x is None:
                return None
            x = x[k]
        return x
    else:
        return x[key]


def get_closest_item_with_tuple_key(x, key):
    if not isinstance(key, tuple):
        key = (key,)

    for k in key:
        x_type = check_type(x)
        if x_type.minor == 'dict' and k in x:
            x = x[k]
        elif x_type.minor == 'list' and k < len(x):
            x = x[k]
        elif x_type.major == 'container':
            return None
        else:
            return x
    return x


def set_item_with_tuple_key(x, key, value):
    if isinstance(key, tuple):
        for k in key[:-1]:
            x = x[k]
        x[key[-1]] = value
    else:
        x[key] = value


def new_container(k):
    if type(k) is int:
        x = []
    else:
        x = {}

    return x


def insert_tupled_key(x, k, v, default=None):
    if x is None and default is None:
        x = new_container(k[0])
    elif x is None:
        x = default

    xi = x

    for ki, kip1 in zip(k[:-1], k[1:]):

        if isinstance(xi, list):
            assert type(ki) is int and len(xi) == ki, 'Invalid key'
            xi.append(new_container(kip1))

        elif ki not in xi:
            xi[ki] = new_container(kip1)

        xi = xi[ki]

    ki = k[-1]
    if isinstance(xi, list):
        assert type(ki) is int and len(xi) == ki, 'Invalid key'
        xi.append(v)
    else:
        xi[ki] = v

    return x


def build_container_from_tupled_keys(keys, values):
    keys = sorted(keys)

    x = None
    for ki, vi in zip(keys, values):
        x = insert_tupled_key(x, ki, vi)

    return x


def tqdm_beam(x, *args, threshold=10, stats_period=1, message_func=None, enable=None, notebook=True, **argv):
    """
    Beam's wrapper for the tqdm progress bar. It features a universal interface for both jupyter notebooks and .py files.
    In addition, it provides a "lazy progress bar initialization". The progress bar is initialized only if its estimated
    duration is longer than a threshold.

    Parameters
    ----------
        x:
        threshold : float
            The smallest expected duration (in Seconds) to generate a progress bar. This feature is used only if enable
            is set to None.
        stats_period: float
            The initial time period (in seconds) to calculate the ineration statistics (iters/sec). This statistics is used to estimate the expected duction of the entire iteration.
        message_func: func
            A dynamic message to add to the progress bar. For example, this message can plot the instantaneous loss.
        enable: boolean/None
            Whether to enable the progress bar, disable it or when set to None, use lazy progress bar.
        notebook: boolean
            A boolean that overrides the internal calculation of is_notebook. Set to False when you want to avoid printing notebook styled tqdm bars (for example, due to multiprocessing).
    """

    my_tqdm = tqdm_notebook if (is_notebook() and notebook) else tqdm

    if enable is False:
        for xi in x:
            yield xi

    elif enable is True:

        pb = my_tqdm(x, *args, **argv)
        for xi in pb:
            if message_func is not None:
                pb.set_description(message_func(xi))
            yield xi

    else:

        iter_x = iter(x)

        if 'total' in argv:
            l = argv['total']
            argv.pop('total')
        else:
            try:
                l = len(x)
            except TypeError:
                l = None

        t0 = timer()

        stats_period = stats_period if l is not None else threshold
        n = 0
        while (te := timer()) - t0 <= stats_period:
            n += 1
            try:
                yield next(iter_x)
            except StopIteration:
                return

        long_iter = None
        if l is not None:
            long_iter = (te - t0) / n * l > threshold

        if l is None or long_iter:
            pb = my_tqdm(iter_x, *args, initial=n, total=l, **argv)
            for xi in pb:
                if message_func is not None:
                    pb.set_description(message_func(xi))
                yield xi
        else:
            for xi in iter_x:
                yield xi


def get_edit_ratio(s1, s2):
    import Levenshtein as lev
    return lev.ratio(s1, s2)


def get_edit_distance(s1, s2):
    import Levenshtein as lev
    return lev.distance(s1, s2)


def filter_dict(d, keys):
    if keys is True:
        return d

    if keys is False:
        return {}

    keys_type = check_type(keys)

    if keys_type.major == 'scalar':
        keys = [keys]

    elif keys_type.minor in ['list', 'tuple']:
        keys = set(keys)
    else:
        raise ValueError(f"keys must be a scalar, list or tuple. Got {keys_type}")

    return {k: v for k, v in d.items() if k in keys}


def none_function(*args, **kwargs):
    return None


class NoneClass:
    def __init__(self, *args, **kwargs):
        pass
    def __getattr__(self, item):
        return none_function


class NullClass:
    pass


def beam_traceback(exc_type=None, exc_value=None, tb=None, context=3):

    if exc_type is None:
        exc_type, exc_value, tb = sys.exc_info()

    if exc_type is None:
        print("No exception found, Printing stack only:")
        return f"{traceback.print_stack()}"
    else:
        return jupyter_like_traceback(exc_type=exc_type, exc_value=exc_value, tb=tb, context=context)


def jupyter_like_traceback(exc_type=None, exc_value=None, tb=None, context=3):

    if exc_type is None:
        exc_type, exc_value, tb = sys.exc_info()

    # Extract regular traceback
    tb_list = traceback.extract_tb(tb)

    # Generate context for each traceback line
    extended_tb = []
    for frame in tb_list:
        filename, lineno, name, _ = frame
        start_line = max(1, lineno - context)
        lines = linecache.getlines(filename)[start_line - 1: lineno + context]
        for offset, line in enumerate(lines, start_line):
            marker = '---->' if offset == lineno else ''
            extended_tb.append(f"{filename}({offset}): {marker} {line.strip()}")

    # Combine the context with the error message
    traceback_text = '\n'.join(extended_tb)
    return f"{traceback_text}\n{exc_type.__name__}: {exc_value}"


def retry(func=None, retrials=3, logger=None, name=None, verbose=False, sleep=1):
    if func is None:
        return partial(retry, retrials=retrials, sleep=sleep)

    name = name if name is not None else func.__name__
    @wraps(func)
    def wrapper(*args, **kwargs):
        local_retrials = retrials
        last_exception = None
        while local_retrials > 0:
            try:
                return func(*args, **kwargs)
            except KeyboardInterrupt as e:
                raise e
            except Exception as e:
                last_exception = e
                local_retrials -= 1
                if logger is not None:

                    if local_retrials == np.inf:
                        retry_message = f"Retrying {name}..."
                    else:
                        retry_message = f"Retries {local_retrials}/{retrials} left."

                    logger.warning(f"Exception occurred in {name}. {retry_message}")

                    if verbose:
                        logger.warning(jupyter_like_traceback())

                if local_retrials > 0:
                    time.sleep(sleep * (1 + np.random.rand()))
        if last_exception:
            raise last_exception

    return wrapper


def run_forever(func=None, *args, sleep=1, name=None, logger=None, **kwargs):
    return retry(func=func, *args, retrials=np.inf, logger=logger, name=name, sleep=sleep, **kwargs)


def parse_text_to_protocol(text, protocol='json'):

    if protocol == 'json':
        import json
        res = json.loads(text)
    elif protocol == 'html':
        from bs4 import BeautifulSoup

        res = BeautifulSoup(text, 'html.parser')
    elif protocol == 'xml':
        from lxml import etree

        res = etree.fromstring(text)
    elif protocol == 'csv':
        import pandas as pd
        from io import StringIO

        res = pd.read_csv(StringIO(text))
    elif protocol == 'yaml':
        import yaml
        res = yaml.load(text, Loader=yaml.FullLoader)

    elif protocol == 'toml':
        import toml
        res = toml.loads(text)

    else:
        raise ValueError(f"Unknown protocol: {protocol}")

    return res


class Slicer:
    def __init__(self, x, x_type=None, wrap_object=False):
        self.x = x
        if x_type is None:
            x_type = check_type(x)
        self.x_type = x_type
        self.wrap_object = wrap_object

    def __getitem__(self, item):
        return slice_array(self.x, item, x_type=self.x_type, wrap_object=self.wrap_object)


class DataObject:
    def __init__(self, data, data_type=None):
        self.data = data
        self._data_type = data_type

    @property
    def data_type(self):
        if self._data_type is None:
            self._data_type = check_type(self.data)
        return self._data_type


def slice_array(x, index, x_type=None, indices_type=None, wrap_object=False):

    if x_type is None:
        x_type = check_minor_type(x)
    else:
        x_type = x_type.minor

    if indices_type is None:
        indices_type = check_minor_type(index)
    else:
        indices_type = indices_type.minor

    if indices_type == 'pandas':
        index = index.values
    if indices_type == 'other': # the case where there is a scalar value with a dtype attribute
        index = int(index)
    if x_type == 'numpy':
        return x[index]
    elif x_type == 'pandas':
        return x.iloc[index]
    elif x_type == 'tensor':
        if x.is_sparse:
            x = x.to_dense()
        return x[index]
    elif x_type == 'list':
        return [x[i] for i in index]
    else:
        try:
            xi = x[index]
            if wrap_object:
                xi = DataObject(xi)
            return xi
        except:
            raise TypeError(f"Cannot slice object of type {x_type}")


def is_arange(x, convert_str=True):
    x_type = check_type(x)

    if x_type.element in ['array', 'object', 'empty', 'none', 'unknown']:
        return False

    if convert_str and x_type.element == 'str':
        pattern = re.compile(r'^(?P<prefix>.*?)(?P<number>\d+)(?P<suffix>.*?)$')
        df = []
        for xi in x:
            match = pattern.match(xi)
            if match:
                df.append(match.groupdict())
            else:
                return None, False
        df = pd.DataFrame(df)
        if not df['prefix'].nunique() == 1 or not df['suffix'].nunique() == 1:
            return None, False

        arr_x = df['number'].astype(int).values
    else:
        arr_x = np.array(x)

    try:
        arr_x = arr_x.astype(int)
        argsort = np.argsort(arr_x)
        arr_x = arr_x[argsort]
    except (ValueError, TypeError):
        return None, False

    isa = np.issubdtype(arr_x.dtype, np.number) and (np.abs(np.arange(len(x)) - arr_x).sum() == 0)

    if not isa:
        argsort = None

    return argsort, isa


# convert a dict to list if is_arange is True
def dict_to_list(x, convert_str=True):
    x_type = check_type(x)

    if not x:
        return []

    if x_type.minor != 'dict':
        return x

    keys = np.array(list(x.keys()))
    argsort, isa = is_arange(keys, convert_str=convert_str)

    if isa:
        return [x[k] for k in keys[argsort]]
    else:
        return x


class Timer(object):
    def __init__(self, logger, name='', silence=False, timeout=None, task=None, task_args=None, task_kwargs=None):
        self.name = name
        self.logger = logger
        self.silence = silence
        self.timeout = timeout
        self.task = task
        self.task_args = task_args or ()
        self.task_kwargs = task_kwargs or {}
        self._elapsed = 0
        self.paused = True
        self.t0 = None
        self.executor = None
        self.future = None

    def __enter__(self):
        if not self.silence:
            self.logger.info(f"Starting timer: {self.name}")
        self.run()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        elapsed = self.pause()
        if not self.silence:
            self.logger.info(f"Timer {self.name} paused. Elapsed time: {pretty_format_number(elapsed)} Sec")

    def reset(self):
        self._elapsed = 0
        self.paused = True
        self.t0 = None
        return self

    @property
    def elapsed(self):
        if self.paused:
            return self._elapsed
        return self._elapsed + time.time() - self.t0

    def pause(self):
        if self.paused:
            return self._elapsed
        self._elapsed = self._elapsed + time.time() - self.t0
        self.paused = True
        return self._elapsed

    def run(self):
        self.paused = False
        self.t0 = time.time()

        if self.task is not None:

            self.logger.info(f"Starting task with timeout of {self.timeout} seconds.")
            self.executor = ThreadPoolExecutor(max_workers=1)
            self.future = self.executor.submit(self.task, *self.task_args, **self.task_kwargs)

            try:
                if self.future:
                    self.future.result(timeout=self.timeout)
            except TimeoutError:
                self.logger.info(f"Timer {self.name} exceeded timeout of {self.timeout} seconds.")
            finally:
                elapsed = self.pause()
                if not self.silence:
                    self.logger.info(f"Timer {self.name} paused. Elapsed time: {elapsed} Sec")
                if self.executor:
                    self.executor.shutdown()

    def __str__(self):
        return f"Timer {self.name}: state: {'paused' if self.paused else 'running'}, elapsed: {self.elapsed}"

    def __repr__(self):
        return str(self)


class ThreadSafeDict(dict):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.lock = threading.Lock()

    def __setitem__(self, key, value):
        with self.lock:
            super().__setitem__(key, value)

    def __getitem__(self, key):
        with self.lock:
            return super().__getitem__(key)

    def __delitem__(self, key):
        with self.lock:
            super().__delitem__(key)

    def update(self, *args, **kwargs):
        with self.lock:
            super().update(*args, **kwargs)

    def pop(self, key, *args):
        with self.lock:
            return super().pop(key, *args)

    def clear(self):
        with self.lock:
            super().clear()

    def setdefault(self, key, default=None):
        with self.lock:
            return super().setdefault(key, default)

    def popitem(self):
        with self.lock:
            return super().popitem()

    def copy(self):
        with self.lock:
            return super().copy()

    def keys(self):
        with self.lock:
            return list(super().keys())

    def values(self):
        with self.lock:
            return list(super().values())

    def items(self):
        with self.lock:
            return list(super().items())


def mixin_dictionaries(*dicts):
    res = {}
    for d in dicts[::-1]:
        if d is not None:
            res.update(d)
    return res


def get_class_properties(cls):
    properties = []
    for attr_name in dir(cls):
        attr_value = getattr(cls, attr_name)
        if isinstance(attr_value, property):
            properties.append(attr_name)
    return properties


def pretty_print_dict(d, name):

    # Convert each key-value pair to 'k=v' format and join them with commas
    formatted_str = ", ".join(f"{k}={v}" for k, v in d.items())

    # Enclose in parentheses
    formatted_str = f"{name}({formatted_str})"
    return formatted_str


import copy
import torch.multiprocessing as mp
import inspect
import traceback
import time
import os

from ..utils import (set_seed, is_notebook,)
from ..path import beam_path
from ..logger import beam_logger as logger
from ..config import get_beam_llm, BeamConfig


def training_closure(rank, world_size, experiment, alg, **kwargs):

    if not rank:
        alg.training_closure(**kwargs)
        checkpoint_file = experiment.checkpoints_dir.joinpath(f'checkpoint_{alg.epoch + 1:06d}')
        alg.save_checkpoint(checkpoint_file)


def worker_executor(experiment, alg, algorithm_generator, manager=None, dataset=None, alg_args=None, alg_kwargs=None,
                    dataset_args=None, dataset_kwargs=None, **kwargs):

    if manager is None:
        rank = 0
        world_size = 1
        manager = {}
    else:
        rank = manager.rank
        world_size = manager.world_size

    if rank == 0:
        manager['is_done'] = False

    experiment.set_rank(rank, world_size, devices=manager.devices)
    set_seed(seed=experiment.hparams.seed, constant=rank + 1, increment=False,
             deterministic=experiment.hparams.deterministic)

    alg = algorithm_generator(experiment, alg, dataset=dataset, alg_args=alg_args, alg_kwargs=alg_kwargs,
                                   dataset_args=dataset_args, dataset_kwargs=dataset_kwargs, rank=rank)

    if rank == 0:
        experiment.writer_control()
    results = {}

    try:
        for i, results in enumerate(iter(alg)):

            if manager['is_done']:
                break

            experiment.save_model_results(copy.deepcopy(results), alg, i)

        if rank == 0:
            logger.info(f"Training is done, Worker terminates.")

    except KeyboardInterrupt as e:

        if rank == 0:
            logger.warning(f"KeyboardInterrupt: Training was interrupted, Worker terminates.")
            logger.debug(f"KeyboardInterrupt: {e}")
            training_closure(rank, world_size, experiment, alg)

    except Exception as e:

        tb = traceback.format_exc()

        llm = get_beam_llm() if experiment.llm is None else experiment.llm

        if llm is not None:
            explain = llm.explain_traceback(tb)
            logger.error(f"LLM Message: {explain}")

        if rank == 0:

            logger.error(f"Exception: {e}")
            logger.error(f"Exception: {tb}")
            logger.error(f"Exception: Training was interrupted, Worker terminates, but checkpoint will be saved.")
            training_closure(rank, world_size, experiment, alg)

        if not is_notebook():
            raise e

    if hasattr(manager, 'cleanup'):
        manager.cleanup()
    experiment.writer_cleanup()
    manager['is_done'] = True

    if rank == 0:
        return alg, results
    return None, None



import atexit
import inspect
import io
import pickle
import time
from collections import defaultdict
from queue import Queue, Empty
from threading import Thread
from uuid import uuid4 as uuid

from ..logger import beam_logger as logger
from ..utils import find_port
from ..config import to_dict
from ..core import MetaDispatcher

try:
    import torch
    has_torch = True
except ImportError:
    has_torch = False


class BeamServer(MetaDispatcher):

    def __init__(self, obj, *args, use_torch=True, batch=None, max_wait_time=1.0, max_batch_size=10,
                 tls=False, n_threads=4, application=None, predefined_attributes=None, **kwargs):

        super().__init__(obj, *args, asynchronous=False, **kwargs)

        if use_torch and has_torch:
            self.load_function = torch.load
            self.dump_function = torch.save
            self.serialization_method = 'torch'
        else:
            self.load_function = pickle.load
            self.dump_function = pickle.dump
            self.serialization_method = 'pickle'

        self.max_wait_time = max_wait_time
        self.max_batch_size = max_batch_size
        self.tls = tls
        self.n_threads = n_threads

        self._request_queue = None
        self._response_queue = None
        self.application = application

        if batch:

            if type(batch) is bool:
                self.batch = ['predict', '__call__']
            elif type(batch) is str:
                self.batch = [batch]
            elif type(batch) is list:
                self.batch = batch
            else:
                raise ValueError(f"Unknown batch type: {batch}")

            # Initialize and start batch inference thread
            self.centralized_thread = Thread(target=self._centralized_batch_executor)
            self.centralized_thread.daemon = True
            self.centralized_thread.start()
        else:
            self.centralized_thread = None
            self.batch = []

        atexit.register(self._cleanup)

        if predefined_attributes is None:
            predefined_attributes = {}
        self._predefined_attributes = predefined_attributes

    def set_variable(self, client, name, value, *args, **kwargs):

        if client == 'beam':
            value = self.load_function(value)

        setattr(self.obj, name, value)
        return {'success': True}

    def get_variable(self, client, name):

        logger.info(f"Getting variable: {name}")
        value = getattr(self.obj, name)
        logger.info(f"value: {value}")

        if client == 'beam':
            io_results = io.BytesIO()
            self.dump_function(value, io_results)
            io_results.seek(0)
            return io_results

        return value

    def _cleanup(self):
        if self.centralized_thread is not None:
            self.centralized_thread.join()

    @property
    def request_queue(self):
        if self._request_queue is None:
            self._request_queue = Queue()
        return self._request_queue

    @property
    def response_queue(self):
        if self._response_queue is None:
            self._response_queue = defaultdict(Queue)
        return self._response_queue

    @classmethod
    def build_algorithm_from_path(cls, path, alg, override_hparams=None, dataset=None, alg_args=None, alg_kwargs=None,
                             dataset_args=None, dataset_kwargs=None, **argv):

        from ..experiment import Experiment
        experiment = Experiment.reload_from_path(path, override_hparams=override_hparams, **argv)
        alg = experiment.algorithm_generator(alg, dataset=dataset, alg_args=alg_args, alg_kwargs=alg_kwargs,
                                                  dataset_args=dataset_args, dataset_kwargs=dataset_kwargs)
        return cls(alg)

    def run_non_blocking(self, **kwargs):
        return self.run(non_blocking=True, **kwargs)

    def run(self, non_blocking=False, **kwargs):
        if non_blocking:
            run_thread = Thread(target=self.run_thread, kwargs=kwargs)
            run_thread.daemon = True
            run_thread.start()
        else:
            run_thread = self.run_thread(**kwargs)
        return run_thread

    def run_thread(self, host=None, port=None, **kwargs):
        if host is None:
            host = "0.0.0.0"
        port = find_port(port=port, get_port_from_beam_port_range=True, application=self.application)
        logger.info(f"Opening a {self.application} inference serve on port: {port}")
        return self._run(host=host, port=port, **kwargs)

    def _run(self, host="0.0.0.0", port=None, **kwargs):
        raise NotImplementedError

    def _centralized_batch_executor(self):

        from ..data import BeamData
        while True:
            logger.info(f"Starting a new batch inference")
            batch = []

            while len(batch) < self.max_batch_size:

                if len(batch) == 1:
                    start_time = time.time()
                    elapsed_time = 0
                elif len(batch) > 1:
                    elapsed_time = time.time() - start_time
                    if elapsed_time > self.max_wait_time:
                        logger.info(f"Max wait time reached, moving to execution")
                        break

                try:
                    if len(batch) > 0:
                        logger.info(f"Waiting for task, for {self.max_wait_time-elapsed_time} seconds")
                        task = self.request_queue.get(timeout=self.max_wait_time-elapsed_time)
                    else:
                        logger.info(f"Waiting for task")
                        task = self.request_queue.get()
                    batch.append(task)
                    logger.info(f"Got task with req_id: {task['req_id']}")
                except Empty:
                    logger.info(f"Empty queue, moving to execution")
                    break

            if len(batch) > 0:
                logger.info(f"Executing batch of size: {len(batch)}")

                methods = defaultdict(list)
                for task in batch:
                    methods[task['method']].append(task)

                for method, tasks in methods.items():

                    logger.info(f"Executing method: {method} with {len(tasks)} tasks")
                    func = getattr(self.obj, method)

                    # currently we support only batching with a single argument
                    data = {task['req_id']: task['args'][0] for task in tasks}

                    is_beam_data = type(data) is BeamData

                    bd = BeamData.simple(data)
                    bd = bd.apply(func)

                    if is_beam_data:
                        results = {task['req_id']: bd[task['req_id']] for task in tasks}
                    else:
                        results = {task['req_id']: bd[task['req_id']].values for task in tasks}

                    for req_id, result in results.items():

                        logger.info(f"Putting result for task: {req_id}")
                        self.response_queue[req_id].put(result)

    @property
    def metadata(self):
        return None

    def get_info(self):

        obj = self.real_object

        d = {'name': None, 'obj': self.type, 'serialization': self.serialization_method}

        if obj is None:
            d['attributes'] = self._predefined_attributes.copy()
            d['hparams'] = None
            d['vars_args'] = None

        elif self.type == 'function':
            d['vars_args'] = obj.__code__.co_varnames
        else:
            d['vars_args'] = obj.__init__.__code__.co_varnames
            if hasattr(obj, 'hparams'):
                d['hparams'] = to_dict(obj.hparams)
            else:
                d['hparams'] = None

            attributes = self._predefined_attributes.copy()
            for name, attr in inspect.getmembers(obj):
                if type(name) is not str:
                    continue
                if not name.startswith('_') and (inspect.ismethod(attr) or inspect.isfunction(attr)):
                    attributes[name] = 'method'
                elif not name.startswith('_') and not inspect.isbuiltin(attr):
                    attributes[name] = 'variable'

            properties = inspect.getmembers(type(obj), lambda m: isinstance(m, property))
            for name, attr in properties:
                if not name.startswith('_'):
                    attributes[name] = 'property'

            d['attributes'] = attributes

        if hasattr(obj, 'name'):
            d['name'] = obj.name

        if hasattr(self, 'metadata'):
            metadata = self.metadata
            if metadata is not None:
                d['metadata'] = self.metadata

        return d

    def batched_query_algorithm(self, method, args, kwargs):

        # Generate a unique request ID
        req_id = str(uuid())
        response_queue = self.response_queue[req_id]

        logger.info(f"Putting task with req_id: {req_id}")
        self.request_queue.put({'req_id': req_id, 'method': method, 'args': args, 'kwargs': kwargs})

        # Wait for the result
        result = response_queue.get()

        logger.info(f"Got result for task with req_id: {req_id}")
        del self.response_queue[req_id]
        return result

    def call(self, client, *args, **kwargs):

        if self.type == 'function':
            return self.query_algorithm(client, None, *args, **kwargs)
        else:
            return self.query_algorithm(client, '__call__', *args, **kwargs)

    def query_algorithm(self, client, method, args, kwargs, return_raw_results=False):

        if client == 'beam':
            args = self.load_function(args)
            kwargs = self.load_function(kwargs)

        if method in self.batch:
            results = self.batched_query_algorithm(method, args, kwargs)
        else:
            if method is None:
                method = self.obj
            else:
                method = getattr(self.obj, method)
            results = method(*args, **kwargs)

        if not return_raw_results and client == 'beam':
            io_results = io.BytesIO()
            self.dump_function(results, io_results)
            io_results.seek(0)
            return io_results

        return results



from typing import Container
import datetime
from optuna import Study
from optuna.trial import TrialState, FrozenTrial

from .optuna import OptunaBase
from ..core import Processor
from .params import HPOConfig
from ..utils import lazy_property


class FiniteMemoryStudy(Study):

    def __init__(self, *args, replay_buffer_size=None, time_window=None,  **kwargs):
        super().__init__(*args, **kwargs)
        assert replay_buffer_size is not None or time_window is not None, \
            "Either replay_buffer_size or time_window must be specified"
        self.replay_buffer_size = replay_buffer_size
        self.time_window = time_window

    def _get_trials(
        self,
        deepcopy: bool = True,
        states: Container[TrialState] | None = None,
        use_cache: bool = False,
    ) -> list[FrozenTrial]:

        trials = super()._get_trials(deepcopy=deepcopy, states=states, use_cache=use_cache)

        if self.replay_buffer_size is not None:
            trials = trials[-self.replay_buffer_size:]
        elif self.time_window is not None:
            trials = [trial for trial in trials if trial.datetime_start > datetime.datetime.now() - self.time_window]
        else:
            raise ValueError("Either replay_buffer_size or time_window must be specified")

        return trials


class LifelongHPO(Processor, OptunaBase):

    def __init__(self, *args, **kwargs):

        hpo_config = HPOConfig(*args, **kwargs)
        super().__init__(hparams=hpo_config)
        self.time_window = self.hparams.get('time_window')
        self.replay_buffer_size = self.hparams.get('replay_buffer_size')
        self.direction = self.hparams.get('mode')




    @lazy_property
    def sampler(self):
        return self.get_sampler()

    @lazy_property
    def study(self) -> FiniteMemoryStudy:
        return FiniteMemoryStudy(storage=self.storage, sampler=self.sampler, pruner=self.pruner,
                                 study_name=self.name, direction=self.direction,
                                 load_study=self.load_study, replay_buffer_size=self.replay_buffer_size,
                                 time_window=self.time_window)


from ..path import BeamURL
from ..utils import beam_base_port, beam_service_port


def get_broker_url(broker=None, broker_username=None, broker_password=None, broker_port=None, broker_scheme=None,
                   broker_host=None):

    default_ports = {'amqp': 5672}
    if broker_scheme is None:
        broker_scheme = 'amqp'
    if broker_host is None:
        broker_host = 'localhost'
        if broker_port is None:
            beam_ports = {'amqp': beam_service_port('RABBITMQ_PORT')}
            broker_port = beam_ports.get(broker_scheme, None)

    if broker_port is None:
        broker_port = default_ports.get(broker_scheme, None)

    broker_url = BeamURL(url=broker, username=broker_username, password=broker_password, port=broker_port,
                              scheme=broker_scheme, hostname=broker_host)

    return broker_url


def get_backend_url(backend=None, backend_username=None, backend_password=None, backend_port=None, backend_scheme=None,
                    backend_host=None):

    default_ports = {'amqp': 5672, 'redis': 6379}
    if backend_scheme is None:
        backend_scheme = 'redis'
    if backend_host is None:
        backend_host = 'localhost'
        if backend_port is None:
            beam_ports = {'amqp': beam_service_port('RABBITMQ_PORT'), 'redis': beam_service_port('REDIS_PORT')}
            backend_port = beam_ports.get(backend_scheme, None)

    if backend_port is None:
        backend_port = default_ports.get(backend_scheme, None)

    backend_url = BeamURL(url=backend, username=backend_username, password=backend_password, port=backend_port,
                              scheme=backend_scheme, hostname=backend_host)
    return backend_url



from functools import partial

from .utils import get_broker_url, get_backend_url
from ..utils import lazy_property
from ..core import MetaAsyncResult, MetaDispatcher


class CeleryAsyncResult(MetaAsyncResult):

    @classmethod
    def from_str(cls, value, app=None):
        return cls(app.AsyncResult(value))

    @property
    def value(self):
        if self._value is None and self.is_ready:
            self._value = self.obj.get()  # Timeout can be adjusted
        return self._value

    def wait(self, timeout=None):

        try:
            return self.obj.get(timeout=timeout)
        except self.obj.TimeoutError:
            return None

    @property
    def hex(self):
        return self.obj.task_id

    @property
    def is_ready(self):
        if self._is_ready is None:
            self._is_ready = self.obj.ready()
        return self._is_ready

    @property
    def is_success(self):
        return self.obj.successful()

    @property
    def state(self):
        return self.obj.state

    @property
    def args(self):
        return self.obj.args

    @property
    def kwargs(self):
        return self.obj.kwargs

    def __repr__(self):
        return f"CeleryAsyncResult({self.hex}, is_ready={self.is_ready}, is_success={self.is_success})"


class CeleryDispatcher(MetaDispatcher):

    def __init__(self, *args, name=None, broker=None, backend=None,
                 broker_username=None, broker_password=None, broker_port=None, broker_scheme=None, broker_host=None,
                 backend_username=None, backend_password=None, backend_port=None, backend_scheme=None,
                 backend_host=None, asynchronous=True, log_level='INFO', **kwargs):

        # in celery obj is not used
        super().__init__(None, *args, name=name, asynchronous=asynchronous, **kwargs)

        self.broker_url = get_broker_url(broker=broker, broker_username=broker_username,
                                         broker_password=broker_password, broker_port=broker_port,
                                         broker_scheme=broker_scheme, broker_host=broker_host)

        self.backend_url = self.backend_url = get_backend_url(backend=backend, backend_username=backend_username,
                                           backend_password=backend_password, backend_port=backend_port,
                                           backend_scheme=backend_scheme, backend_host=backend_host)

        self.log_level = log_level

    @lazy_property
    def broker(self):
        from celery import Celery
        app = Celery(self.name, broker=self.broker_url.url, backend=self.backend_url.url)
        app.conf.update(
            worker_log_level=self.log_level,
            broker_connection_retry_on_startup=True
        )
        return app

    def __call__(self, *args, **kwargs):
        return self.dispatch('function', *args, **kwargs)

    def poll(self, task_id, timeout=0):
        async_res = CeleryAsyncResult.from_str(task_id, app=self.broker)
        return async_res.wait(timeout=timeout)

    def metadata(self, task_id, *args, **kwargs):
        res = self.broker.AsyncResult(task_id)
        d = {'task_id': task_id, 'state': res.state, 'result': res.result,
             'traceback': res.traceback if res.state == 'FAILURE' else None, 'status': res.status,
             'children': res.children, 'retries': res.retries, "parent_id": res.parent.id if res.parent else None,
             'exception': str(res.result) if res.state == 'FAILURE' else None,
             'date_done': res.date_done if hasattr(res, 'date_done') else None,
             'runtime': res.runtime if hasattr(res, 'runtime') else None}

        return d

    def dispatch(self, attribute, *args, **kwargs):
        res = self.broker.send_task(attribute, args=args, kwargs=kwargs)
        res = CeleryAsyncResult(res)
        if self.asynchronous:
            return res
        else:
            return res.value

    def getattr(self, item):
        return partial(self.dispatch, item)




import math

import torch
import faiss
import numpy as np

from ..data import BeamData
from ..logger import beam_logger as logger
from ..path import beam_path, local_copy
from ..utils import pretty_format_number, lazy_property, as_numpy, check_type, as_tensor, beam_device
from .core import BeamSimilarity, Similarities


class DenseSimilarity(BeamSimilarity):

    def __init__(self, *args, vector_store=None, d=None, expected_population=int(1e6),
                 metric='l2', training_device='cpu', inference_device='cpu', ram_footprint=2**8*int(1e9),
                 gpu_footprint=24*int(1e9), exact=False, nlists=None, faiss_M=None,
                 reducer='umap', **kwargs):

        '''
        To Choose an index, follow https://github.com/facebookresearch/faiss/wiki/Guidelines-to-choose-an-index
        @param d:
        @param expected_population:
        @param metric:
        @param ram_size:
        @param gpu_size:
        @param exact_results:
        @param reducer:
        '''

        super().__init__(*args, vector_dimension=d, expected_population=expected_population,
                         metric=metric, training_device=training_device,
                         inference_device=inference_device, ram_footprint=ram_footprint,
                         gpu_footprint=gpu_footprint, exact=exact, nlists=nlists, M=faiss_M,
                         reducer=reducer, **kwargs)

        d = self.get_hparam('vector_dimension', d)
        expected_population = self.get_hparam('expected_population', expected_population)
        metric = self.get_hparam('metric', metric)
        training_device = str(beam_device(self.get_hparam('training_device', training_device)))
        inference_device = beam_device(self.get_hparam('inference_device', inference_device))

        if inference_device.type == 'cpu':
            inference_device = inference_device.type
        else:
            inference_device = inference_device.index

        ram_footprint = self.get_hparam('ram_footprint', ram_footprint)
        gpu_footprint = self.get_hparam('gpu_footprint', gpu_footprint)
        exact = self.get_hparam('exact', exact)
        nlists = self.get_hparam('nlists', nlists)
        faiss_M = self.get_hparam('faiss_M', faiss_M)
        reducer = self.get_hparam('reducer', reducer)

        metrics = {'l2': faiss.METRIC_L2, 'l1': faiss.METRIC_L1, 'linf': faiss.METRIC_Linf,
                   'cosine': faiss.METRIC_INNER_PRODUCT, 'ip': faiss.METRIC_INNER_PRODUCT,
                   'js': faiss.METRIC_JensenShannon}
        metric = metrics[metric]
        self.normalize = False
        if metric == 'cosine':
            self.normalize = True

        # choosing nlists: https://github.com/facebookresearch/faiss/issues/112,
        #  https://github.com/facebookresearch/faiss/wiki/Guidelines-to-choose-an-index
        if nlists is None:
            if expected_population <= int(1e6):
                # You will need between 30*K and 256*K vectors for training (the more the better)
                nlists = int(8 * math.sqrt(expected_population))
            elif expected_population > int(1e6) and expected_population <= int(1e7):
                nlists = 2 ** 16
            elif expected_population > int(1e7) and expected_population <= int(1e8):
                nlists = 2 ** 18
            else:
                nlists = 2 ** 20

        if vector_store is None:
            if inference_device == 'cpu':

                if exact:
                    logger.info(f"Using Flat Index. Expected RAM footprint is "
                                f"{pretty_format_number(4 * d * expected_population / int(1e6))} MB")
                    vector_store = faiss.IndexFlat(d, metric)
                else:
                    if faiss_M is None:
                        faiss_M = 2 ** np.arange(2, 7)[::-1]
                        footprints = (d * 4 + faiss_M * 8) * expected_population
                        M_ind = np.where(footprints < ram_footprint)[0]
                        if len(M_ind):
                            faiss_M = int(faiss_M[M_ind[0]])
                    if faiss_M is not None:
                        logger.info(f"Using HNSW{faiss_M}. Expected RAM footprint is "
                                    f"{pretty_format_number(footprints[M_ind[0]] / int(1e6))} MB")
                        vector_store = faiss.IndexHNSWFlat(d, faiss_M, metric)
                    else:
                        logger.info(f"Using OPQ16_64,IVF{nlists},PQ8 Index")
                        vector_store = faiss.index_factory(d, f'OPQ16_64,IVF{nlists},PQ8')

            else:

                res = faiss.StandardGpuResources()
                if exact:
                    config = faiss.GpuIndexFlatConfig()
                    config.device = inference_device
                    logger.info(f"Using GPUFlat Index. Expected GPU-RAM footprint is "
                                f"{pretty_format_number(4 * d * expected_population / int(1e6))} MB")

                    vector_store = faiss.GpuIndexFlat(res, d, metric, config)
                else:

                    if (4 * d + 8) * expected_population <= gpu_footprint:
                        logger.info(f"Using GPUIndexIVFFlat Index. Expected GPU-RAM footprint is "
                                    f"{pretty_format_number((4 * d + 8) * expected_population / int(1e6))} MB")
                        config = faiss.GpuIndexIVFFlatConfig()
                        config.device = inference_device
                        vector_store = faiss.GpuIndexIVFFlat(res, d, nlists, faiss.METRIC_L2, config)
                    else:

                        if faiss_M is None:
                            faiss_M = 2 ** np.arange(2, 7)[::-1]
                            footprints = (faiss_M + 8) * expected_population
                            M_ind = np.where(footprints < gpu_footprint)[0]
                            if len(M_ind):
                                faiss_M = faiss_M[M_ind[0]]
                                faiss_M = faiss_M[M_ind[0]]
                        if faiss_M is not None:
                            logger.info(f"Using GPUIndexIVFFlat Index. Expected GPU-RAM footprint is "
                                        f"{pretty_format_number((faiss_M + 8) * expected_population / int(1e6))} MB")

                            config = faiss.GpuIndexIVFPQConfig()
                            config.device = inference_device
                            vector_store = faiss.GpuIndexIVFPQ(res, d, nlists, faiss_M, 8, faiss.METRIC_L2, config)
                        else:
                            logger.info(f"Using OPQ16_64,IVF{nlists},PQ8 Index")
                            vector_store = faiss.index_factory(d, f'OPQ16_64,IVF{nlists},PQ8')
                            vector_store = faiss.index_cpu_to_gpu(res, inference_device, vector_store)

        if vector_store is None:
            logger.error("Cannot find suitable index type")
            raise Exception

        self.vector_store = vector_store
        self.index = None
        if vector_store.ntotal and self.index is None:
            self.index = BeamData(data=torch.arange(vector_store.ntotal, device=inference_device))

        self.inference_device = inference_device

        self.training_vs = None
        res = faiss.StandardGpuResources()
        if training_device != 'cpu' and inference_device == 'cpu':
            self.training_vs = faiss.index_cpu_to_gpu(res, training_device, vector_store)

        self.training_device = training_device
        self.reducer_type = reducer

    @lazy_property
    def reducer(self):
        if self.reducer_type == 'umap':
            import umap
            reducer = umap.UMAP()
        elif self.reducer_type == 'tsne':
            from sklearn.manifold import TSNE
            reducer = TSNE()
        else:
            raise NotImplementedError
        return reducer

    def train(self, x):

        x = as_numpy(x)
        self.vector_store.train(x)

    @property
    def is_trained(self):
        return self.vector_store.is_trained

    def add(self, x, index=None, train=False):

        x, index = self.extract_data_and_index(x, index)
        self.add_index(x, index)

        if (not self.is_trained) or train:
            self.train(x)

        self.vector_store.add(x)

    def search(self, x, k=1):

        x_type = check_type(x)
        device = x.device if x_type.minor == 'tensor' else None

        x, _ = self.extract_data_and_index(x)
        D, I = self.vector_store.search(x, k)

        if self.index is not None:
            I = self.index[I]
        if x_type.minor == 'tensor':
            I = as_tensor(I, device=device)
            D = as_tensor(D, device=device)

        return Similarities(index=I, distance=D, model='faiss', metric=self.metric)

    def reduce(self, z):
        return self.reducer.fit_transform(z)

    @property
    def exclude_pickle_attributes(self):
        return ['index', 'vector_store', 'training_vs']

    def save_state(self, path, ext=None):

        path = beam_path(path)
        path.mkdir()

        path.joinpath('index.npy').write(self.index)
        with local_copy(path.joinpath('vectore_store.bin'), as_beam_path=False) as p:
            faiss.write_index(self.vector_store, p)
        if self.training_vs:
            with local_copy(path.joinpath('training_vs.bin'), as_beam_path=False) as p:
                faiss.write_index(self.training_vs, p)

    def load_state(self, path):

        path = beam_path(path)

        self.index = path.joinpath('index.npy').read()
        with local_copy(path.joinpath('vectore_store.bin'), as_beam_path=False) as p:
            self.vector_store = faiss.read_index(p)
        if path.joinpath('training_vs.bin').is_file():
            with local_copy(path.joinpath('training_vs.bin'), as_beam_path=False) as p:
                self.training_vs = faiss.read_index(p)



import argparse
import warnings

import numpy as np
import pandas as pd
import torch

from .sampler import UniversalBatchSampler
from ..utils import (recursive_batch, to_device, recursive_device, container_len, beam_device, as_tensor, check_type,
                     as_numpy, slice_to_index, DataBatch, lazy_property)

from ..path import beam_path
from ..data import BeamData


class UniversalDataset(torch.utils.data.Dataset):

    def __init__(self, *args, index=None, label=None, device=None, target_device=None, to_torch=True,
                 index_mapping='backward', **kwargs):
        """
        Universal Beam dataset class

        @param args:
        @param index:
        @param device:
        @param target_device: if not None, the dataset is responsible to transform samples into this dataset.
        This is useful when we want to transform a sample to the GPU during the getitem routine in order to speed-up the
        computation.
        @param kwargs:
        """
        super().__init__()

        if device is None:
            device = 'cpu'
        device = beam_device(device)
        target_device = beam_device(target_device)

        self.index = None
        self.set_index(index, mapping=index_mapping)

        if not hasattr(self, 'indices_split'):
            self.indices = {}
        if not hasattr(self, 'labels_split'):
            self.labels_split = {}
        if not hasattr(self, 'probs'):
            self.probs = {}

        # The training label is to be used when one wants to apply some data transformations/augmentations
        # only in training mode
        self.training = False
        self.statistics = None
        self._target_device = target_device
        self.to_torch = to_torch

        if len(args) >= 1 and isinstance(args[0], argparse.Namespace):
            self.hparams = args[0]
            args = args[1:]

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if len(args) == 1:
                d = args[0]
                if isinstance(d, dict):
                    self.data = {k: as_tensor(v, device=device) for k, v in d.items()}
                    self.data_type = 'dict'
                elif isinstance(d, list) or isinstance(d, tuple):
                    self.data = [as_tensor(v, device=device) for v in d]
                    self.data_type = 'list'
                else:
                    self.data = d
                    self.data_type = 'simple'
            elif len(args):
                self.data = [as_tensor(v, device=device) for v in args]
                self.data_type = 'list'
            elif len(kwargs):
                self.data = {k: as_tensor(v, device=device) for k, v in kwargs.items()}
                self.data_type = 'dict'
            else:
                self.data = None

        self.label = as_tensor(label, device=self.device)

    @lazy_property
    def target_device(self):
        if self._target_device is not None:
            return self._target_device

        if hasattr(self, 'hparams'):
            if self.hparams.get('accelerate', False) and self.hparams.get('device_placement', False):
                return None
            if self.hparams.get('device', None) is not None and self.hparams.get('n_gpus', 1) <= 1:
                return beam_device(self.hparams.get('device', None))

        return None

    def set_index(self, index, mapping='backward'):

        self.index = None
        if index is not None:
            index_type = check_type(index)
            if index_type.minor == 'tensor':
                index = as_numpy(index)
            if mapping == 'backward':
                index = pd.Series(data=np.arange(len(index)), index=index)
                # check if index is not a simple arange
                if np.abs(index.index.values - np.arange(len(index))).sum() > 0:
                    self.index = index
            elif mapping == 'forward':
                index = pd.Series(data=index, index=np.arange(len(index)))
                self.index = index
            else:
                raise NotImplementedError(f"Mapping type: {mapping} not supported")


    def train(self):
        self.training = True

    def eval(self):
        self.training = False

    @lazy_property
    def data_type(self):
        return check_type(self.data).minor

    def getitem(self, ind):

        if self.data_type == 'dict':

            ind_type = check_type(ind, check_minor=False)
            if ind_type.element == 'str':
                if ind_type.major == 'scalar':
                    return self.data[ind]
                return [self.data[k] for k in ind]

            return {k: recursive_batch(v, ind) for k, v in self.data.items()}

        elif self.data_type == 'list':
            return [recursive_batch(v, ind) for v in self.data]
        elif self.data_type == 'simple':
            return self.data[ind]
        else:
            return self.data[ind]

    @classmethod
    def get_subset(cls, self, subset):
        index = self.indices[subset]
        return cls(self.data, index=index, label=self.label, device=self.device,
                     target_device=self.target_device, to_torch=self.to_torch, index_mapping='forward')

    def __getitem__(self, ind):

        if type(ind) is str:
            return UniversalDataset.get_subset(self, ind)

        if self.index is not None:

            ind = slice_to_index(ind, l=self.index.index.max()+1)

            ind_type = check_type(ind, check_element=False)
            if ind_type.minor == 'tensor':
                loc = as_numpy(ind)
            else:
                loc = ind
                ind = as_tensor(ind)

            if ind_type.major == 'scalar':
                loc = [loc]

            iloc = self.index.loc[loc].values

        else:

            ind = slice_to_index(ind, l=len(self))
            iloc = ind

        sample = self.getitem(iloc)
        if self.to_torch:
            sample = as_tensor(sample, device=self.target_device)
        elif self.target_device is not None:
            sample = to_device(sample, device=self.target_device)

        label = None
        if self.label is not None:
            label = self.label[iloc]

        return DataBatch(index=ind, data=sample, label=label)

    @lazy_property
    def device(self):

        if self.data_type == 'dict':
            device = recursive_device(next(iter(self.data.values())))
        elif self.data_type == 'list':
            device = recursive_device(self.data[0])
        elif self.data_type == 'simple':
            device = self.data.device
        elif hasattr(self.data, 'device') and self.data.device is not None:
            device = self.data.device
        else:
            device = 'cpu'

        return beam_device(device)

    def __repr__(self):
        return repr(self.data)

    @property
    def values(self):
        return self.data

    # def save(self, path):
    #
    #     bd_path = beam_path(path).joinpath('beam_data')
    #     bd = BeamData(self.data, index=self.index, label=self.label, path=bd_path, device=self.device)
    #     bd.store()
    #
    #     state = self.state
    #     if has_beam_ds and isinstance(state, BeamData):
    #         state.store(path=path)
    #     else:
    #         path = beam_path(path)
    #         path.write(state)

    def __len__(self):

        if self.index is not None:
            return len(self.index)

        if self.data_type is None:
            self.data_type = check_type(self.data).minor

        if self.data_type == 'dict':
            return container_len(next(iter(self.data.values())))
        elif self.data_type == 'list':
            return container_len(self.data[0])
        elif self.data_type == 'simple':
            return len(self.data)
        elif hasattr(self.data, '__len__'):
            return len(self.data)
        else:
            raise NotImplementedError(f"For data type: {type(self.data)}")

    def split(self, validation=None, test=None, seed=5782, stratify=False, labels=None,
                    test_split_method='uniform', time_index=None, window=None):
        """
                partition the data into train/validation/split folds.
                Parameters
                ----------
                validation : float/int/array/tensor
                    If float, the ratio of the data to be used for validation. If int, should represent the total number of
                    validation samples. If array or tensor, the elements are the indices for the validation part of the data
                test :  float/int/array/tensor
                   If float, the ratio of the data to be used for test. If int, should represent the total number of
                   test samples. If array or tensor, the elements are the indices for the test part of the data
                seed : int
                    The random seed passed to sklearn's train_test_split function to ensure reproducibility. Passing seed=None
                    will produce randomized results.
                stratify: bool
                    If True, and labels is not None, partition the data such that the distribution of the labels in each part
                    is the same as the distribution of the labels in the whole dataset.
                labels: iterable
                    The corresponding ground truth for the examples in data
                """

        from sklearn.model_selection import train_test_split

        if labels is None:
            labels = self.label
        if self.label is None:
            self.label = labels

        indices = np.arange(len(self))
        if time_index is None:
            time_index = indices

        if test is None:
            pass
        elif check_type(test).major == 'array':
            self.indices['test'] = as_tensor(test, dtype=torch.long)
            indices = np.sort(list(set(indices).difference(set(as_numpy(test)))))

            if labels is not None:
                self.labels_split['test'] = labels[self.indices['test']]
                # labels = labels[indices]

        elif test_split_method == 'uniform':

            if labels is not None:
                labels_to_split = labels[indices]
                indices, test, _, self.labels_split['test'] = train_test_split(indices, labels_to_split,
                                                                               random_state=seed,
                                                                               test_size=test,
                                                                               stratify=labels_to_split if stratify else None)
            else:
                indices, test = train_test_split(indices, random_state=seed, test_size=test)

            self.indices['test'] = as_tensor(test, dtype=torch.long)
            if seed is not None:
                seed = seed + 1

        elif test_split_method == 'time_based':
            ind_sort = np.argsort(time_index)
            indices = indices[ind_sort]

            test_size = int(test * len(self)) if type(test) is float else test
            self.indices['test'] = as_tensor(indices[-test_size:], dtype=torch.long)
            indices = indices[:-test_size]

            if labels is not None:
                labels = labels[ind_sort]
                self.labels_split['test'] = labels[self.indices['test']]

        if validation is None:
            pass
        elif check_type(validation).major == 'array':
            self.indices['validation'] = as_tensor(validation, dtype=torch.long)
            indices = np.sort(list(set(indices).difference(set(as_numpy(validation)))))

            if labels is not None:
                self.labels_split['validation'] = labels[self.indices['validation']]

        else:
            if type(validation) is float:
                validation = len(self) / len(indices) * validation

            if labels is not None:

                labels_to_split = labels[indices]
                indices, validation, _, self.labels_split['validation'] = train_test_split(indices, labels_to_split, random_state=seed,
                                                                                                test_size=validation, stratify=labels_to_split if stratify else None)
            else:
                indices, validation = train_test_split(indices, random_state=seed, test_size=validation)

            self.indices['validation'] = as_tensor(validation, dtype=torch.long)

        self.indices['train'] = as_tensor(indices, dtype=torch.long)
        if labels is not None:
            self.labels_split['train'] = labels[indices]

    def set_statistics(self, stats):
        self.statistics = stats

    def build_sampler(self, batch_size, subset=None, persistent=True, oversample=False, weight_factor=1., expansion_size=int(1e7),
                       dynamic=False, buffer_size=None, probs_normalization='sum', tail=True, sample_size=100000):

        from sklearn.utils.class_weight import compute_sample_weight

        if subset is None:
            if self.index is not None:
                indices = self.index.index.values
            else:
                indices = torch.arange(len(self))
        else:
            indices = self.indices[subset]

        if not persistent:
            return UniversalBatchSampler(indices, batch_size, shuffle=False,
                                         tail=tail, once=True, dynamic=False)

        probs = None
        if oversample and subset in self.labels_split and self.labels_split[subset] is not None:
            probs = compute_sample_weight('balanced', y=self.labels_split[subset]) ** weight_factor
            probs_normalization = 'sum'
        elif subset is None and check_type(self.probs).major == 'array':
            probs = self.probs
        elif subset in self.probs:
            probs = self.probs[subset]

        return UniversalBatchSampler(indices,
                                     batch_size, probs=probs, shuffle=True, tail=tail,
                                     once=False, expansion_size=expansion_size,
                                     dynamic=dynamic, buffer_size=buffer_size,
                                     probs_normalization=probs_normalization,
                                     sample_size=sample_size, device=self.device)

    def build_dataloader(self, sampler, num_workers=0, pin_memory=None, timeout=0, collate_fn=None,
                   worker_init_fn=None, multiprocessing_context=None, generator=None, prefetch_factor=2):

        kwargs = {}
        if num_workers > 0:
            kwargs['prefetch_factor'] = prefetch_factor

        try:
            d = self.device.type if self.target_device is None else self.target_device
            pin_memory_ = ('cpu' == d)
        except NotImplementedError:
            pin_memory_ = True

        if pin_memory is None:
            pin_memory = pin_memory_
        else:
            pin_memory = pin_memory and pin_memory_

        persistent_workers = (num_workers > 0 and not sampler.once)

        return torch.utils.data.DataLoader(self, sampler=sampler, batch_size=None,
                                             num_workers=num_workers, pin_memory=pin_memory, timeout=timeout,
                                             worker_init_fn=worker_init_fn, collate_fn=collate_fn,
                                             multiprocessing_context=multiprocessing_context, generator=generator,
                                             persistent_workers=persistent_workers, **kwargs)
